/////////////////////////////////////////////////////////////////////
// $Id: Commands.cpp,v 1.251 2005/08/24 21:07:45 klas Exp $
//
// Usage: Ftp commands and data transfer functions
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define COMMANDS_CPP
#define NETINCLS
#define BABY_CORE
#include "Commands.h"

class Commands comm;

void *data_wrapper(void *void_data)
{
  class Data *data = static_cast<class Data*>(void_data);
  string message;
  time_t start_time = time(NULL);
  size_t trans_size = 0; // will hold the total number of sent bytes

  data->thread_id(pthread_self());

  pthread_setspecific(my_data, void_data);
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  data->host_name(inet_ntoa(data->addr().sin_addr));

  // I don't like it, but we need to drop permissions again...
  //   not true for new threads, need to look deeper into this.
  struct passwd pwd, *pwd_ptr;
  char pwdata[PWBUF_SIZE];
  getpwnam_r(data->user->user_name().c_str(), &pwd, pwdata, sizeof(pwdata),
      &pwd_ptr);
  uid_t uid = pwd.pw_uid;
  gid_t gid = pwd.pw_gid;
  setuid(uid);
  setgid(gid);

#ifdef USE_TLS // handshakes etc
  if(data->settings(TLS_PROT_D))
    if(!data->engage_tls()) // failed handshake
      message =
        "451 Failed to negotiate TLS as was asked. Closing data connection.";
#endif // USE_TLS
  while(message.empty())
  {
    char *read_buf = new char[data->packet_size()];
    size_t size = 0, size2 = 0;
    if(data->settings(RETR_TRANS))
    {
      size2 = read(data->file_id(), read_buf, data->packet_size());
      if(size2 > 0)
        size = data->send(read_buf, size2);
    }
    else // STOR_STRANS
      size = data->receive(read_buf, data->packet_size() - 1);

    trans_size += size;
    if(!data->settings(RETR_TRANS))
    {
      if(size != 0 && (write(data->file_id(), read_buf, size) == -1))
      {
        logging->log_this(3, TYPE_DEBUG, "Write failed to file: " +
            data->file_name() + " from host: " + data->host_name() +
            " with error %m");
        message = "451 File transfer failed.";
      }
      else if(size == 0)
      {
        if(close(data->file_id()) != 0)
          message = "550 Failed saving file.";
        else
          message = "226 Transfer complete.";
      }
    }
    else // RETR_TRANS
    {
      if(size != size2)
        message = "550 Transfer failed.";
      else if(size2 != data->packet_size())
        message = "226 Transfer complete.";
    }
    delete[] read_buf;
  }

  *data->user << message << "\r\n";
  if(config->logger.xferlog)
  {
    string time_string;
    util.time_string(time_string, "%a %b %d %H:%M:%S %Y ");

    // as ' ' separates fields in the xferlog we convert them to '_' here
    string temp_name = data->file_name();
    for(string::iterator name_it = temp_name.begin();
        name_it != temp_name.end(); name_it++)
      if(isspace(*name_it))
        *name_it = '_';

    // let us build a string!
    string xferlog = time_string;
    xferlog += util.itos(time(NULL) - start_time);
    xferlog += " " + data->host_name();
    xferlog += " " + util.itos(trans_size);
    xferlog += " " + temp_name + " ";
    xferlog += (data->settings(BINARY)?"b":"a");
    xferlog += " _ "; // we don't support any special-action to flag here.
    xferlog += (data->settings(RETR_TRANS)?"o":"i");
    xferlog += (data->user->settings(ANONYMOUS)? // a for anonymous, r for real
        " a " + data->user->anonymous_id(): // if the user is anonymous we log
        " r " + data->user->user_name()); // the id string, else the user name.
    xferlog += " FTP"; // maybe ftps for encrypted? no. IANA.
    xferlog += " 0"; // there should be more auth methods available here?
    xferlog += " *"; // or should this be uid?

    logging->log_this(1, TYPE_XFER, xferlog);
  }

  pthread_exit(NULL);
}

// parsing strings so we get something good to work with,
// beware of this function, it is one of the most expensive
// functions in the game...
string Commands::parsedir(User *user, string orig_string)
{
  string new_string;

  if(orig_string[0] != '/')
    orig_string = user->current_dir() + '/' + orig_string;

  class StringTokenizer *token = new StringTokenizer(orig_string, "/", false);
  vector<string> dir_array;
  dir_array.reserve(token->countTokens());

  while(token->hasMoreTokens())
  {
    dir_array.push_back(token->nextToken());

    if(dir_array.back().size() < 3)
    { // to save us from expensive string comparisons
      if(dir_array.back() == "..")
      {
        dir_array.pop_back();
        if(!dir_array.empty())
          dir_array.pop_back();
      }
      else if(dir_array.back() == ".")
        dir_array.pop_back();
    }
  }
  delete token;

  if(dir_array.empty())
    new_string = '/';
  else
    for(vector<string>::iterator dir_it = dir_array.begin();
        dir_it != dir_array.end(); dir_it++)
      new_string += '/' + *dir_it;

  if(user->root_dir().size() > 1)
    if(new_string.find(user->root_dir()) != 0)
      new_string = user->root_dir() + new_string;


  return(new_string);
}

int pam_conversate(int num_msg, const struct pam_message **msg,
    struct pam_response **resp, void *appdata_ptr)
{
  struct pam_response *reply = NULL;
  int count = 0;
  struct pam_data *auth_data = (static_cast<struct pam_data*>(appdata_ptr));

  for(int i = 0; i < num_msg; i++)
  {
    struct pam_response *reply_tmp;
    switch(msg[i]->msg_style)
    {
      case PAM_PROMPT_ECHO_ON:
      case PAM_PROMPT_ECHO_OFF:
        reply_tmp = (static_cast<struct pam_response*>
            (malloc(sizeof(struct pam_response) * (count + 1))));
        memcpy(reply_tmp, reply, count);
        if(reply)
          free(reply);
        reply = reply_tmp;

        // is this really how it should be done?
        if(msg[i]->msg_style == PAM_PROMPT_ECHO_ON)
          reply[i].resp = strndup(auth_data->user_name.c_str(),
              auth_data->user_name.size());
        else
          reply[i].resp = strndup(auth_data->password.c_str(),
              auth_data->password.size());

        reply[count++].resp_retcode = PAM_SUCCESS;
        break;

      case PAM_TEXT_INFO:
        break;

      case PAM_ERROR_MSG:
      default:
        if(reply)
          free(reply);
        return PAM_CONV_ERR;
    }
  }
  *resp = reply;
  return PAM_SUCCESS;
}
