/////////////////////////////////////////////////////////////////////
// $Id: Commands.h,v 1.96 2005/04/29 17:58:16 klas Exp $
// Usage: Ftp commands and data transfer functions
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#ifndef COMMANDS_H
#define COMMANDS_H

#if (defined(COMMANDS_CPP) || defined(COMMANDS2_CPP))
#include "generic.h"
#include "Utilities.h"
#include "Configuration.h"
#include "Log.h"
#include "Connection.h"
#include "Data.h"
#include "User.h"
#include "Handler.h"
#include "StringTokenizer.h"
#endif

#define BABY_COM(com)  bool com(User *user, string &argument, unsigned int flag)

class Commands
{
  private:
    string parsedir(User*, string);
    string mlsx(User*, string);
    string list_file(string);
    string mlst_string(User*);
  public:
    BABY_COM(user);
    BABY_COM(pass);
    BABY_COM(retr);
    BABY_COM(stor);
    BABY_COM(port);
    BABY_COM(quit);
    BABY_COM(list);
    BABY_COM(help);
    BABY_COM(mkd);
    BABY_COM(rmd);
    BABY_COM(cwd);
    BABY_COM(dele);
    BABY_COM(pasv);
    BABY_COM(rname);
    BABY_COM(opts);
    BABY_COM(rest);
    BABY_COM(abor);
    // functions for simple commands that don't justify an own function
    BABY_COM(sim1);
    BABY_COM(sim2);
    sem_t pam_lock; // pam is not threadsafe.
#ifdef USE_TLS
    // collection of encryption commands
    BABY_COM(auth);
#endif // USE_TLS
};

extern class Commands comm;

// the extra flags we can play with...
const int COM_LONG = (1 << 0);  // for list()
const int COM_SHORT = (1 << 1);
const int COM_MLSD = (1 << 2);

const int COM_FROM = (1 << 0); // for rname()
const int COM_TO = (1 << 1);

const int COM_STOR = (1 << 0); // for stor()
const int COM_APPE = (1 << 1);

const int COM_CDUP = (1 << 0);
const int COM_CWD  = (1 << 1);

const int COM_NOOP = (1 << 0); // for sim1()
const int COM_PWD  = (1 << 1);
const int COM_SYST = (1 << 2);
const int COM_FEAT = (1 << 3);
const int COM_TYPE = (1 << 4);
const int COM_MODE = (1 << 5);
const int COM_STRU = (1 << 6);
#ifdef USE_TLS
const int COM_PBSZ = (1 << 7);
const int COM_PROT = (1 << 8);
const int COM_CCC  = (1 << 9);
const int COM_CONF = (1 << 10);
const int COM_MIC = (1 << 11);
const int COM_ENC = (1 << 12);
#endif // USE_TLS

const int COM_SIZE = (1 << 0); // for sim2()
const int COM_MDTM = (1 << 1);
const int COM_MLST = (1 << 2);

const int COM_PORT = (1 << 0); // port and pasv
const int COM_EPRT = (1 << 1);
const int COM_PASV = (1 << 2);
const int COM_EPSV = (1 << 3);

// remember that these are used in conunctions with those above..
const int BEFORE_USER = (1 << 31);

const int PROT_IPV4 = 1;// EPRT and EPSV
const int PROT_IPV6 = 2;


void *data_wrapper(void*) __attribute__ ((noreturn));

extern vector<User*> user_list;

#ifdef USE_TLS
extern gnutls_certificate_credentials x509_cred;
extern int tls_db_store(void*, gnutls_datum, gnutls_datum);
extern gnutls_datum tls_db_fetch(void*, gnutls_datum);
extern int tls_db_delete(void*, gnutls_datum);
#endif // USE_TLS

int pam_conversate(int, const struct pam_message**,
    struct pam_response**, void*);

struct pam_data {
  string user_name;
  string password;
};

#endif
