/////////////////////////////////////////////////////////////////////
// $Id: Commands2.cpp,v 1.169 2005/08/24 21:07:45 klas Exp $
//
// Usage: Ftp commands and data transfer functions
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define COMMANDS2_CPP
#define NETINCLS
#define BABY_CORE
#include "Commands.h"

BABY_COM(Commands::quit)
{
  *user << "221 Goodbye. You transfered a total of "
    << user->transfered() << " bytes." << "\r\n";

  user->settings(FAILURE, B_SET);
  return(true);
}

BABY_COM(Commands::user)
{
  string message;
  bool ret = true, bad = false;
  if(user->logged_in() == LOGGED_IN)
  {
    message = "503 You are already logged in!";
    ret = false;
  }
  else
  {
    int current_connections = 0;
    pthread_rwlock_rdlock(&user_list_lock);
    for(vector<User*>::iterator user_it = user_list.begin();
        user_it != user_list.end(); user_it++)
      if((*user_it)->user_name() == argument)
        current_connections++;
    pthread_rwlock_unlock(&user_list_lock);

    user->user_name(argument);
    user->logged_in(USER_SUPPLIED);

    if(config->get_user(argument).noftp == true)
    {
      message = "421 Sorry, you are not allowed to use ftp. disconnecting.";
      bad = true;
    }
    else if(current_connections >= config->get_user(argument).user_limit){
      message =
        "421 You have reached your maximum number of control connections.";
      bad = true;
    }
#ifdef USE_TLS
    else if(config->get_user(argument).encryption == DENY &&
        user->settings(TLS_ENABLED))
    {
      message =  "421 Sorry, you are not allowed to connect with"
        "encryption enabled, disconnecting.";
      bad = true;
    }
    else if(user->settings(TLS_ENABLED) && user->verify_cert())
    {
      message = "230 Logged in as " + argument + ".";
      user->login();
    }
    else if(config->get_user(argument).encryption == FORCE_CERT)
    {
      message = "421 You must supply a valid certificate to login.";
      bad = true;
    }
    else if(config->get_user(argument).encryption == FORCE &&
        !user->settings(TLS_ENABLED))
    {
      message =
        "421 You must enable encryption before logging in, disconnecting.";
      bad = true;
    }
#endif // USE_TLS
    else if(config->general.anonymous >= A_ENABLE &&
        user->user_name() == "anonymous")
      message = "331 Anonymous login. Please, provide your email as password.";
    else if(config->general.anonymous == A_ONLY)
    {
      message = "421 All logins except anonymous disabled, disconnecting.";
      bad = true;
    }
    else
      message = "331 Password required for " + argument + ".";
  }
  *user << message << "\r\n";
  if(bad)
  {
    user->settings(FAILURE, B_SET);
    ret = false;
  }
  return(ret);
}

BABY_COM(Commands::abor)
{
  string message = "226 ";
  if(user->data_vec.empty())
    message += "No Data Connections to cancel";
  else
  {
    for(vector<class Data*>::iterator data_it = user->data_vec.begin();
        data_it != user->data_vec.end(); data_it++)
      if(*data_it)
        pthread_cancel((*data_it)->thread_id());
    message += "All Data Connections Successfully closed.";
  }
  *user << message << ".\r\n";
  return(true);
}

BABY_COM(Commands::pass)
{
  // before everything else, lets check if it is really time for pass...
  if(user->logged_in() != USER_SUPPLIED)
  {
    string message;
    if(user->logged_in() == NOT_LOGGED_IN)
      message = "503 Login with 'USER' first.";
    else if(user->logged_in() == LOGGED_IN)
      message = "503 You are already logged in!";

    *user << message << "\r\n";
    return(false);
  }

  // is anonymous login enabled?
  if(config->general.anonymous >= A_ENABLE &&
      user->user_name() == "anonymous")
  {
    char pwdbuf[PWBUF_SIZE];
    struct passwd pwd, *pwd_ptr;
    getpwuid_r(config->general.user, &pwd, pwdbuf, sizeof(pwdbuf), &pwd_ptr);
    user->user_name(pwd.pw_name);

    // maybe we should do some rudimentary checking of the adress? no?
    logging->log_this(3, TYPE_INFO, "anonymous login from " + user->host_name()
        + " with email " + argument + ".");
    *user << "230 Anonymous login successful.\r\n";
    user->login();
    user->settings(ANONYMOUS, B_SET);
    user->anonymous_id(argument);
    return(true);
  }

  seteuid(0); // we need to become root to check the user
  // pam stuff
  struct pam_data *auth_data = new struct pam_data;
  struct pam_conv pamconv = { &pam_conversate, auth_data };
  int pam_error = 0;
  pam_handle_t *tmp_pamh = user->pamh();

  auth_data->user_name = user->user_name();
  auth_data->password = argument;
  argument.clear();

  sem_wait(&comm.pam_lock);
  pam_error = pam_start("babyftpd", user->user_name().c_str(),
      &pamconv, &tmp_pamh);
  user->pamh(tmp_pamh);

  if(pam_error == 0)
  {
    pam_error = pam_authenticate(user->pamh(), 0);
    if(pam_error == 0)
    {
      pam_error = pam_acct_mgmt(user->pamh(), 0);
      if(pam_error == 0)
        pam_error = pam_setcred(user->pamh(), PAM_ESTABLISH_CRED);
    }
  }
  if(pam_error != 0)
  {
    pam_end(user->pamh(), pam_error);
    sem_post(&pam_lock);
    seteuid(config->general.user); // and back to some other user..
  }

  delete auth_data;

  if(pam_error == 0)
  {
    pam_open_session(user->pamh(), PAM_SILENT);
    user->settings(PAM_DONE, B_SET);

    const char *pam_username;
    pam_get_item(user->pamh(), PAM_USER,
        (reinterpret_cast<const void**>(&pam_username)));
    sem_post(&pam_lock);
    if(user->user_name() != pam_username)
    {
      logging->log_this(3, TYPE_INFO, "PAM changed username of " +
          user->user_name() + " to " + pam_username);
      user->user_name(pam_username);
    }

      *user << "230 Password accepted, proceeding.\r\n";
    user->login();
    return(true);
  }
  *user << "530 Login denied.\r\n";
  return(false);
}

BABY_COM(Commands::help)
{
  string message;
  if(!argument.size())
  {
    int cmd = 0;
    message = "214-The following commands are understood:\r\n ";
    for(hash_map<const char*, struct command_info, hash<const char*>,
        Utilities::eqstr>::iterator cmd_it = handler->cmd_info.begin();
        cmd_it != handler->cmd_info.end(); cmd_it++)
    {
      if(!(*(cmd_it)).second.hlp_info.empty())
      {
        if(cmd != 0 && cmd % 8 == 0)
          message += "\r\n ";
        message += (*(cmd_it)).first;
        message += '\t';
        cmd++;
      }
    }
    message +=
      "\r\n214 if you have any trouble please contact the administrator at "
      + config->general.admin_email + ".";
  }
  else
  {
    // other help stuff here...
    message  = "214 Syntax: ";
    transform(argument.begin(), argument.end(), argument.begin(),
        (int(*)(int))toupper);
    hash_map<const char*, struct command_info, hash<const char*>, Utilities::
      eqstr>::iterator cmd_it = handler->cmd_info.find(argument.c_str());

    if(cmd_it != handler->cmd_info.end() &&
        !(*(cmd_it)).second.hlp_info.empty())
      message += ((*(cmd_it)).first + (*(cmd_it)).second.hlp_info);
    else
      message = "500 No help available for '" + argument + "'";
  }
  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::cwd)
{
  struct stat stt;
  string message, new_dir;
  DIR *tmp_dir;

  if(util.is_set(flag, COM_CDUP))
    argument = "..";

  if((user->root_dir().size() > 1 && (user->current_dir().size() ==
          user->root_dir().size())) && (argument == ".." || argument == "."))
    new_dir = user->root_dir();
  else
    new_dir = this->parsedir(user, argument);


  lstat(new_dir.c_str(), &stt);

  if(S_ISLNK(stt.st_mode))
  {
    char real_name[65] = { 0 };
    int len;

    len = readlink(new_dir.c_str(), real_name, 60);
    real_name[len] = '\0';

    new_dir = this->parsedir(user, real_name);
  }

  if(!(tmp_dir = opendir(new_dir.c_str())))
  {
    if(errno == EACCES)
      message = "550 '" + new_dir + "' permission denied.";
    else
      message = "550 '" + new_dir + "' no such file or directory.";
  }
  else
  {
    if(util.is_set(flag, COM_CDUP))
      message = "200 'CDUP' command successful.";
    else // COM_CWD
      message = "250 'CWD' command successful.";

    user->current_dir(new_dir);
  }
  closedir(tmp_dir);
  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::retr)
{
  struct stat stt;

  string file_name = parsedir(user, argument);
  stat(file_name.c_str(), &stt);
  if(!S_ISREG(stt.st_mode))
  {
    *user << "550 \"" << argument << "\" is not a regular file.\r\n";
    return(false);
  }

  int file_id = open(file_name.c_str(), O_RDONLY);
  if(user->rest_pos() > 0)
  {
    lseek(file_id, user->rest_pos(), SEEK_SET);
    user->rest_pos(0);
  }

  *user << "150 Opening " << (user->settings(BINARY)?"BINARY":"ASCII")
      << " mode data connection for \"" << argument << "\" ("
      << util.itos(stt.st_size) << " bytes).\r\n";

  class Data *data = new Data(user, file_id, file_name, 1);
  if(!data->settings(DATA_OPEN) || user->data_vec_full())
  {
    *user << "425 Can't open data connection.\r\n";
    delete data;
    return(false);
  }

  BABY_WRLOCK(user->lock);
  user->data_vec.push_back(data);
  BABY_UNLOCK(user->lock);
  pthread_t temp_id;
  pthread_create(&temp_id, NULL, data_wrapper,
      static_cast<void*>(data));
  return(true);
}

BABY_COM(Commands::stor)
{
  int file_id;

  string file_name = parsedir(user, argument);

  if(util.is_set(flag, COM_STOR))
    file_id = open(file_name.c_str(), O_WRONLY|O_CREAT|O_TRUNC,
        S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
  else // com_type == COM_APPE
    file_id = open(file_name.c_str(), O_WRONLY|O_APPEND|O_TRUNC,
        S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);

  if(file_id == -1)
  {
    string message;
    if(errno == EEXIST)
      message = "550 File " + file_name + " already exists.";
    else if(errno == EISDIR)
      message = "550 " + file_name + " is a directory.";
    else if(errno == EACCES)
      message = "550 Access to " + file_name + " denied.";
    else
      message = "550 File operation failed.";
    *user << message << "\r\n";
    return false;
  }

  // shouldn't try to restart an appending :p
  if(user->rest_pos() > 0 && util.is_set(flag, COM_STOR))
  {
    lseek(file_id, user->rest_pos(), SEEK_SET);
    user->rest_pos(0);
  }

  *user << "150 Opening " << (user->settings(BINARY)?"BINARY":"ASCII")
      << " mode data connection for \"" << argument << "\".\r\n";

  class Data *data = new Data(user, file_id, file_name, 0);
  if(!data->settings(DATA_OPEN) || user->data_vec_full())
  {
    *user << "425 Can't open data connection.\r\n";
    delete data;
    return(false);
  }

  BABY_WRLOCK(user->lock);
  user->data_vec.push_back(data);
  BABY_UNLOCK(user->lock);
  pthread_t temp_id;
  pthread_create(&temp_id, NULL, data_wrapper,
      static_cast<void*>(data));
  return(true);
}

BABY_COM(Commands::rest)
{
  char *endptr;
  string message;

  off_t fpos = strtoul(argument.c_str(), &endptr, 0);
  if(!*endptr)
  {
    user->rest_pos(fpos);
    message = "350 Restarting at " + argument + ". Send STOR or RETR.";
  }
  else
    message = "501 Restart point not in numeric format.";

  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::list)
{
  struct stat stt;
  string failure;

  if(argument.size() == 0)
    argument = user->current_dir();

  string file = parsedir(user, argument);

  if((lstat(file.c_str(), &stt)) == -1 && ((util.is_set(flag, COM_LONG) ||
          util.is_set(flag, COM_MLSD)) && argument[0] == '-'))
  {
    class StringTokenizer *token = new StringTokenizer(argument, " ", false);
    token->nextToken();
    argument = token->restTokens();
    delete token;
    file = parsedir(user, argument);
    lstat(file.c_str(), &stt);
  }

  if(!S_ISDIR(stt.st_mode) && util.is_set(flag, COM_MLSD))
  {
    *user << "501 Argument to MLSD has to be a directory.\r\n";
    return false;
  }

  class Data *data = new Data(user);

  if(!data->settings(DATA_OPEN))
    failure = "425 Can't open data connection.";
  else
  {
    *user << "150 Opening ";
#ifdef USE_TLS
    if(data->settings(TLS_PROT_D))
    {
      *user << "PROTECTED ASCII mode data connection for file list.\r\n";
      if(!data->engage_tls())
        failure =
          "451 Failed to negotiate TLS as was asked. Closing data connection.";
    }
    else
#endif // USE_TLS
      *user << " ASCII mode data connection for file list.\r\n";
    if(S_ISDIR(stt.st_mode))
    {
      struct dirent dirstr, *dirptr;
      vector<string> dirlist;

      DIR *dirp = opendir(file.c_str());
      while((readdir_r(dirp, &dirstr, &dirptr) == 0) && dirptr != NULL)
      {
#ifdef _DIRENT_HAVE_D_TYPE
        if((dirstr.d_type == DT_DIR) && util.is_set(flag, COM_SHORT))
          continue;
#endif // _DIRENT_HAVE_D_TYPE
        dirlist.push_back(dirstr.d_name);
      }

      closedir(dirp);

      // sorting the directory so we output first things first
      sort(dirlist.begin(), dirlist.end());
      for(vector<string>::iterator dir_it = dirlist.begin();
          dir_it != dirlist.end(); dir_it++)
      {
        if(util.is_set(flag, COM_LONG))
          *data << list_file(file + "/" + *dir_it);
        else if(util.is_set(flag, COM_MLSD))
          *data << mlsx(user, file + "/" + *dir_it);
        else
          *data << *dir_it; // short list.

        *data << "\r\n";
        if(data->settings(FAILURE))
          break;
      }
    }
    else
    {
      if(util.is_set(flag, COM_LONG))
        *data << list_file(file);
      else // NLST
        *data << argument;
      *data << "\r\n";
    }
    if(data->settings(FAILURE))
      failure = "426 LIST transfer failed.";
  }
  if(failure.size() > 0)
    *user << failure << "\r\n";
  else
    *user << "226 Transfer complete.\r\n";

  delete data;
  return(true);
}

string Commands::list_file(string file_name)
{
  char size[20], nlink[6], pwdbuf[PWBUF_SIZE], grpbuf[PWBUF_SIZE];
  string user_name, group_name, time_string;
  struct stat file_stat;
  struct passwd pwd, *pwd_ptr;
  struct group grp, *grp_ptr;

  lstat(file_name.c_str(), &file_stat);

  // number of hard links
  snprintf(nlink, 5, "%3d ", static_cast<int>(file_stat.st_nlink));

  if((getpwuid_r(file_stat.st_uid, &pwd, pwdbuf, sizeof(pwdbuf),
          &pwd_ptr)) == 0)
    user_name = pwd.pw_name;
  else
    user_name = util.itos(file_stat.st_uid);

  if((getgrgid_r(file_stat.st_gid, &grp, grpbuf, sizeof(grpbuf),
          &grp_ptr)) == 0)
    group_name = grp.gr_name;
  else
    group_name = util.itos(file_stat.st_gid);

  snprintf(size, 19, "%9d ", static_cast<int>(file_stat.st_size)); // filesize

  // modification time, either in new ls format, or old..
#ifdef USE_NEWLS
  string time_format = "%Y-%m-%d %H:%M ";
#else
  string time_format;
  // is the file over a year old?
  if((time(NULL) - file_stat.st_mtime) > (60 * 60 * 24 * 365))
    time_format = "%b %e  %Y ";
  else // if it's not, we display the time it was created as well.
    time_format = "%b %e %H:%M ";
#endif // USE_NEWLS

  util.time_string(time_string, time_format, 25, file_stat.st_mtime);

  string message = (util.parse_perms_ls(file_stat.st_mode) + ' ')
    + nlink + user_name + '\t' + group_name + '\t' + size + time_string
    + basename(file_name.c_str()); // parsing everything on one line is neater.

  // if it's a link...
  if(S_ISLNK(file_stat.st_mode))
  {
    char real_name[55];
    int len = readlink(file_name.c_str(), real_name, 50);
    real_name[len] = '\0';
    message += " -> ";
    message += real_name;
  }
  return(message);
}

string Commands::mlsx(User *user, string file_name)
{
  char pwdbuf[PWBUF_SIZE], grpbuf[PWBUF_SIZE];
  struct passwd pwd, *pwd_ptr;
  struct group grp, *grp_ptr;

  struct stat stt;
  string message;

  lstat(file_name.c_str(), &stt);

  // type.
  if(user->settings(FACT_TYPE))
  {
    string type;
    if(S_ISREG(stt.st_mode))
      type = "file";
    else if(S_ISDIR(stt.st_mode))
    {
      if(file_name == user->current_dir())
      {
        type = "cdir";
        file_name += "/.";
      }
      else if(file_name == parsedir(user, user->current_dir() + "/.."))
        type = "pdir";
      else
        type = "dir";
    }
    else if(S_ISLNK(stt.st_mode))
      type = "OS.unix=slink";
    else if(S_ISCHR(stt.st_mode))
      type = "OS.unix=chr";
    else if(S_ISBLK(stt.st_mode))
      type = "OS.unix=blk";
    else if(S_ISFIFO(stt.st_mode))
      type = "OS.unix=fifo"; // or?
    else if(S_ISSOCK(stt.st_mode))
      type = "OS.unix=sock"; // or?

    message =  "Type=" + type + ";";
  }

  if(user->settings(FACT_SIZE))
    message += "Size=" + util.itos(stt.st_size) + ";";
  if(user->settings(FACT_MODIFY))
  {
    string tmp_time;
    util.time_string(tmp_time, "Modify=%Y%m%d%H%M%S;", 23, stt.st_mtime);
    message += tmp_time;
  }
  if(user->settings(FACT_UNIQUE))
    message += "Unique=" + util.itos(stt.st_dev) + "g" +
      util.itos(stt.st_ino) + ";";
  if(user->settings(FACT_PERM))
  {
    bool owner = false, group = false;

    if(getpwnam_r(user->user_name().c_str(), &pwd, pwdbuf, sizeof(pwdbuf),
          &pwd_ptr))
    {
      if(pwd.pw_uid == stt.st_uid)
        owner = true;
      else if(pwd.pw_gid == stt.st_gid)
        group = true;
    }

    if(!group && !owner)
    {
      if(getgrgid_r(stt.st_gid, &grp, grpbuf, sizeof(grpbuf), &grp_ptr))
        for(int i = 0; grp.gr_mem[i] != NULL; i++)
          if(user->user_name() == grp.gr_mem[i])
            group = true;
    }

    message += "Perm=";
    if(owner)
    {
      if(S_ISDIR(stt.st_mode))
      {
        if(util.is_set(stt.st_mode, S_IRUSR)) message += "l";
        if(util.is_set(stt.st_mode, S_IWUSR)) message += "cmp";
        if(util.is_set(stt.st_mode, S_IXUSR)) message += "e";
      }
      if(util.is_set(stt.st_mode, S_IRUSR)) message += "r";
      if(util.is_set(stt.st_mode, S_IWUSR)) message += "adfw";
    }
    else if(group)
    {
      if(S_ISDIR(stt.st_mode))
      {
        if(util.is_set(stt.st_mode, S_IRGRP)) message += "l";
        if(util.is_set(stt.st_mode, S_IWGRP)) message += "cmp";
        if(util.is_set(stt.st_mode, S_IXGRP)) message += "e";
      }
      if(util.is_set(stt.st_mode, S_IRGRP)) message += "r";
      if(util.is_set(stt.st_mode, S_IWGRP)) message += "adfw";
    }
    else
    {
      if(S_ISDIR(stt.st_mode))
      {
        if(util.is_set(stt.st_mode, S_IROTH)) message += "l";
        if(util.is_set(stt.st_mode, S_IWOTH)) message += "cmp";
        if(util.is_set(stt.st_mode, S_IXOTH)) message += "e";
      }
      if(util.is_set(stt.st_mode, S_IROTH)) message += "r";
      if(util.is_set(stt.st_mode, S_IWOTH)) message += "adfw";
    }
    message += ";";
  }
  if(user->settings(FACT_U_MODE))
    message += "Unix.mode=" + util.parse_perms_num(stt.st_mode) + ";";
  if(user->settings(FACT_U_UID))
    message += "Unix.uid=" + util.itos(stt.st_uid) + ";";
  if(user->settings(FACT_U_GID))
    message += "Unix.gid=" + util.itos(stt.st_gid) + ";";
  if(user->settings(FACT_U_LINK))
    message += "Unix.link=" + util.itos(stt.st_nlink) + ";";
  if(user->root_dir().size() > 1)
    file_name.erase(0, user->root_dir().size());
  message += " " + file_name;

  return(message);
}

BABY_COM(Commands::mkd)
{
  string message, new_dir = parsedir(user, argument);

  if(mkdir(new_dir.c_str(), 0755))
  {
    message = "550 ";
    if(errno == EEXIST)
      message += "File exists.";
    else if(errno == EACCES)
      message += "Write access denied.";
    else if(errno == (ENOTDIR || ENOENT))
      message += "Faulty pathname, directory not created.";
    else
      message += "Directory not created.";
  }
  else
    message = "257 \"" + argument + "\" created.";

  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::rmd)
{
  string message, new_dir = parsedir(user, argument);

  if(rmdir(new_dir.c_str()))
  {
    message = "550 ";
    if(errno == EACCES)
      message += "Write access denied.";
    else if(errno == (ENOTDIR || ENOENT))
      message += "Faulty pathname, directory not removed.";
    else
      message += "Directory not removed.";
  }
  else
    message = "250 \"" + argument + "\" removed.";

  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::dele)
{
  string message, new_file = parsedir(user, argument);

  if(unlink(new_file.c_str()))
  {
    message = "550 ";
    if(errno == EPERM || errno == EACCES)
      message += "File access denied, not deleted.";
    else if(errno == EPERM)
      message += "File \"" + new_file + "\" is a directory.";
    else if(errno == ENOENT || errno == EISDIR)
      message += "File not found.";
    else
      message = "450 File not available.";
  }
  else
    message = "250 'DELE' command successful.";

  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::port)
{
  string tmp_adress, message;
  int port = -1;

  if(user->settings(EPSV_ONLY))
    message = "501 Only 'EPSV' may be used after the issuing of 'EPSV ALL'";

  else if((config->get_user(user->user_name()).only_passive))
    message =
      "501 Sorry, active data transfers are disabled for your account.";

  else if(util.is_set(flag, COM_PORT))
  {
    class StringTokenizer *token = new StringTokenizer(argument, ",", false);

    if(token->countTokens() != 6)
      message = "501 Illegal 'PORT' command.";

    else
    {
      tmp_adress = token->nextToken() + ".";  // if we do this on one line
      tmp_adress += token->nextToken() + "."; // it builds the string from
      tmp_adress += token->nextToken() + "."; // the wrong direction, it's
      tmp_adress += token->nextToken();       // weird.

      port = atoi((token->nextToken()).c_str()) * 256 +
          atoi((token->nextToken()).c_str());
    }
    delete token;
  }
  else if(util.is_set(flag, COM_EPRT))
  {
    class StringTokenizer *token =
      new StringTokenizer(argument, argument.substr(0, 1), false);

    if(token->countTokens() != 3 ||
        argument[argument.size() - 1] != argument[0])
      message = "501 Illegal 'EPRT' command.";

    else
    {
      if(token->nextToken() == util.itos(PROT_IPV4))
      {
        tmp_adress = token->nextToken();
        port = atoi(token->nextToken().c_str());
      }
      else
        message = "522 Network protocol not supported, use (1)";
    }
    delete token;
  }

  if(message.empty())
  {
    struct sockaddr_in temp_addr;
    if((inet_aton(tmp_adress.c_str(), &temp_addr.sin_addr)) == 0)
    {
      message = "501 Illegal '";
      message += (util.is_set(flag, COM_PORT) ? "PORT" : "EPRT");
      message += "' command.";
    }
    else
    {
      temp_addr.sin_port = htons(port);
      temp_addr.sin_family = AF_INET;
      user->settings(PASSIVE, B_RESET);
      message += "200 '";
      message += (util.is_set(flag, COM_PORT) ? "PORT" : "EPRT");
      message += "' command successful.";
    }
    user->data_addr(temp_addr);
  }

  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::pasv)
{
  string message;

  if(util.is_set(flag, COM_EPSV) && argument.size() > 0)
  {
    transform(argument.begin(), argument.end(), argument.begin(),
        (int(*)(int))toupper);
    if(argument == "ALL" && util.is_set(flag, COM_EPSV))
    {
      user->settings(EPSV_ONLY, B_SET);
      message = "200 Now only accepting 'EPSV'";
    }
    else
      message = "501 '" + argument + "' is not a valid argument to 'EPSV'.";
  }
  else
  {
    struct sockaddr_in tmp_addr = {0}, tmp_addr1 = {0}, tmp_addr2 = {0};
    socklen_t len = sizeof(tmp_addr);

    if(user->data_id())
      close(user->data_id());

    user->data_id(socket(AF_INET, SOCK_STREAM, IPPROTO_TCP));
    tmp_addr.sin_family = AF_INET;
    tmp_addr.sin_addr.s_addr = INADDR_ANY;
    bind(user->data_id(), reinterpret_cast<struct sockaddr*>
        (&tmp_addr), sizeof(tmp_addr));
    user->data_addr(tmp_addr);
    listen(user->data_id(), 1);
    getsockname(user->data_id(), reinterpret_cast<struct sockaddr*>
        (&tmp_addr1), &len);
    getsockname(user->socket_id(), reinterpret_cast<struct sockaddr*>
        (&tmp_addr2), &len);
    user->settings(PASSIVE, B_SET);

    unsigned int p = ntohs(tmp_addr1.sin_port);

    if(util.is_set(flag, COM_PASV))
    {
      if(user->settings(EPSV_ONLY))
        message =
          "501 Only 'EPSV' may be used after the issuing of 'EPSV ALL'";

      else
      {
        class StringTokenizer *token =
          new StringTokenizer(inet_ntoa(tmp_addr2.sin_addr), ".", false);

        message = "227 Entering Passive Mode (";
        message += token->nextToken() + ",";
        message += token->nextToken() + ",";
        message += token->nextToken() + ",";
        message += token->nextToken() + ",";
        message += util.itos(p / 256) + ","
          + util.itos(p % 256) + ").";

        delete token;
      }
    }
    else if(util.is_set(flag, COM_EPSV))
    {
      if(argument.size() > 0)
        if(argument != util.itos(PROT_IPV4))
          message = "522 Network protocol not supported, use (1)";

      if(message.size() == 0)
        message = "229 Entering Extended Passive Mode (|||" +
          util.itos(p) + "|)";
    }
  }
  *user << message << "\r\n";
  return(true);
}

// rnfr and rnto
BABY_COM(Commands::rname)
{
  string file_name, message;

  file_name = parsedir(user, argument);

  if(util.is_set(flag, COM_FROM))
  {
    user->temp_buf(file_name);
    message = "350 Command OK, waiting for RNTO.";
  }
  else if(util.is_set(flag, COM_TO))
  {
    if(user->temp_buf().size() == 0)
      message = "503 RNTO must come after RNFR.";
    else
    {
      if(!(rename(user->temp_buf().c_str(), file_name.c_str())))
        message = "250 " + user->temp_buf() + " renamed to " + file_name + ".";
      else if(errno == EACCES)
        message = "550 Access denied.";
      else
        message = "553 Rename failed.";
      user->temp_buf("");
    }
  }
  *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::opts)
{
  string command, message;
  class StringTokenizer *st;

  if(argument == "")
  {
    message += "500-Recognized OPTS arguments: \r\n";
    message += " MLST\r\n";
    message += "500 End.";
  }
  else
  {
    transform(argument.begin(), argument.end(), argument.begin(),
        (int(*)(int))toupper);
    st = new StringTokenizer(argument, " ", false);
    command = st->nextToken();
    argument = st->restTokens();
    delete st;

    if(command == "MLST")
    {
      string fact;
      st = new StringTokenizer(argument, ";", false);

      while(st->hasMoreTokens())
      {
        fact = st->nextToken();
        if(fact == "TYPE") user->settings(FACT_TYPE, B_FLIP);
        else if(fact == "SIZE") user->settings(FACT_SIZE, B_FLIP);
        else if(fact == "MODIFY") user->settings(FACT_MODIFY, B_FLIP);
        else if(fact == "UNIQUE") user->settings(FACT_UNIQUE, B_FLIP);
        else if(fact == "PERM") user->settings(FACT_PERM, B_FLIP);
        else if(fact == "UNIX.MODE") user->settings(FACT_U_MODE, B_FLIP);
        else if(fact == "UNIX.UID") user->settings(FACT_U_UID, B_FLIP);
        else if(fact == "UNIX.GID") user->settings(FACT_U_GID, B_FLIP);
        else if(fact == "UNIX.LINK") user->settings(FACT_U_LINK, B_FLIP);
      };

      message = "200 MLST string now showing: " + mlst_string(user);
      delete st;
    }
    else
      message = "501 " + command + " not known as argument to OPTS.";
  }

  *user << message << "\r\n";

  return(true);
}

string Commands::mlst_string(User *user)
{
  string message;

  message += ("Type"); message +=
    (user->settings(FACT_TYPE) ? "*;" : ";");
  message += ("Size"); message +=
    (user->settings(FACT_SIZE) ? "*;" : ";");
  message += ("Modify"); message +=
    (user->settings(FACT_MODIFY) ? "*;" : ";");
  message += ("Unique"); message +=
    (user->settings(FACT_UNIQUE) ? "*;" : ";");
  message += ("Perm"); message +=
    (user->settings(FACT_PERM) ? "*;" : ";");
  message += ("Unix.mode"); message +=
    (user->settings(FACT_U_MODE) ? "*;" : ";");
  message += ("Unix.uid"); message +=
    (user->settings(FACT_U_UID) ? "*;" : ";");
  message += ("Unix.gid"); message +=
    (user->settings(FACT_U_GID) ? "*;" : ";");
  message += ("Unix.link"); message +=
    (user->settings(FACT_U_LINK) ? "*;" : ";");
  return(message);
}

// functions for small left over commands (noop, pwd, etc...)
BABY_COM(Commands::sim1)
{
  string message;

  transform(argument.begin(), argument.end(), argument.begin(),
      (int(*)(int))toupper);

  if(util.is_set(flag, COM_NOOP))
    message = "200 'NOOP': Command successful.";

  else if(util.is_set(flag, COM_PWD))
  {
    string dir = user->current_dir();
    if(user->root_dir().size() > 1)
      dir.erase(0, user->root_dir().size());
    if(dir == "")
      dir = "/";
    message = "257 \"" + dir + "\" is current directory.";
  }

  else if(util.is_set(flag, COM_SYST))
    message = "215 " + config->general.os + " Type: L8";

  else if(util.is_set(flag, COM_FEAT))
  {
    message =  "211-Extensions supported:\r\n";
    message += " SIZE\r\n";
    message += " MDTM\r\n";
    message += " MLST " + mlst_string(user) + "\r\n";
    message += " REST STREAM\r\n";
    message += " EPRT\r\n";
    message += " EPSV\r\n";
#ifdef USE_TLS
    message += " AUTH TLS\r\n";
    message += " PBSZ\r\n";
    message += " PROT\r\n";
    message += " CCC\r\n";
#endif // USE_TLS
    message += "211 End.";
  }

  else if(util.is_set(flag, COM_TYPE))
  {
    if(argument == "A" || argument == "A N")
    {
      message = "200 Type set to " + argument + ".";
      user->settings(BINARY, B_RESET);
    }
    else if(argument == "I" || argument == "L 8")
    {
      message = "200 Type set to " + argument + ".";
      user->settings(BINARY, B_SET);
    }
    else
      message = "504 'TYPE " + argument + "' not understood.";
  }

  else if(util.is_set(flag, COM_MODE))
  {
    if(argument == "S")
    {
      message = "200 Stream MODE enabled.";
    }
    else
      message = "504 'MODE " + argument + "' not understood.";
  }

  else if(util.is_set(flag, COM_STRU))
  {
    if(argument == "F")
      message = "200 Structure set to F.";
    else
      message = "504 'STRU " + argument + "' not understood.";
  }
#ifdef USE_TLS
  else if(util.is_set(flag, COM_PBSZ))
  {
    if(user->settings(TLS_ENABLED))
    {
      if(argument == "0")
      {
        message = "200 PBSZ set for stream transfer.";
        user->settings(TLS_PBSZ, B_SET);
      }
      else
        message = "501 PBSZ has to be set to 0 for use with TLS.";
    }
    else
      message =
        "503 AUTH must be issued and negotiated before PBSZ can be used.";
  }

  else if(util.is_set(flag, COM_PROT))
  {
    if(!user->settings(TLS_ENABLED) || !user->settings(TLS_PBSZ))
      message =
        "503 AUTH and PBSZ has to be issued and negotiated before PROT.";
    else
      switch(argument[0])
      {
        case 'C':
          message = "200 Data connection set to CLEAR level.";
          user->settings(TLS_PROT_D, B_RESET);
          break;
        case 'P':
          message = "200 Data connection set to PRIVATE level.";
          user->settings(TLS_PROT_D, B_SET);
          break;
        case 'S':
        case 'E':
          message = "536 Only CLEAR and PRIVATE level is available with TLS.";
          break;
        default:
          message = "504 Protection level '" + argument + "' not understood.";
      }
  }
  else if(util.is_set(flag, COM_CCC)) // Clear Command Channel
  {
    if(!user->settings(TLS_ENABLED))
      message =
        "533 CCC can only be issued on a protected control connection.";
    else // code for disallowing CCC should be above, containing a 534
    {
      *user << "200 Disabling TLS, control connection is now CLEAR.\r\n";
      user->disengage_tls();
    }
  }
  else if(util.is_set(flag, COM_CONF | COM_MIC | COM_ENC))
  {
    // note, if we implement som other security messure than TLS we might
    // need these commands, for now they reside here and only return
    // a 503 or 537.
    string command;
    if(util.is_set(flag, COM_CONF))
      command = "CONF";
    else if(util.is_set(flag, COM_MIC))
      command += "MIC";
    else
      command += "ENC";

    if(!user->settings(TLS_ENABLED))
      message = "503 Command " + command
        + " can not be issued before a valid security exchange.";
    else
      message = "537 Command " + command + " not used in TLS mode.";
  }
#endif // USE_TLS

  if(message.size() > 0)
    *user << message << "\r\n";
  return(true);
}

BABY_COM(Commands::sim2)
{
  string file_name, message;
  struct stat stt;

  file_name = parsedir(user, argument);
  if(!(stat(file_name.c_str(), &stt)))
  {
    if(util.is_set(flag, COM_SIZE))
      message = "213 " + util.itos(stt.st_size);

    else if(util.is_set(flag, COM_MDTM))
      util.time_string(message, "213 %Y%m%d%H%M%S", 19, stt.st_mtime);

    else if(util.is_set(flag, COM_MLST))
    {
      message  = "250-Listing " + argument + "\r\n";
      message += mlsx(user, file_name) + "\r\n";
      message += "250 End.";
    }
  }
  else
    message = "550 \"" + argument + "\": No such file or directory.";

  *user << message << "\r\n";
  return(true);
}

#ifdef USE_TLS

BABY_COM(Commands::auth)
{
  transform(argument.begin(), argument.end(), argument.begin(),
      (int(*)(int))toupper);

  if(argument == "TLS" || argument == "TLS-C" || argument == "SSL")
  {
    if(config->get_user(user->user_name()).encryption == DENY)
    {
      *user <<
        "421 Sorry, you are not allowed to use encryption, disconnecting.\r\n";
      user->settings(FAILURE, B_SET);
      return false;
    }
    *user << "234 Now starting encryption using TLS.\r\n";
    if(!user->engage_tls())
    { // failure =(
      user->settings(FAILURE, B_SET);
      *user << "421 Handshake failed. Disconnecting.\r\n";
    }
    else
    { // wee!
      if(argument == "SSL")
        user->settings(TLS_PROT_D, B_SET);
    }
  }
  else
    *user << "504 Sorry, I don't understand the mechanism '"
      << argument << "'.\r\n";

  return true;
}
#endif // USE_TLS
