/////////////////////////////////////////////////////////////////////
//  $Id: Configuration.cpp,v 1.101 2005/08/23 19:34:02 klas Exp $
//
//  Usage: Functions and stuff for Configure class, holding and
//           reading config stuff.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file 
//    LICENSE for complete information.                                         
/////////////////////////////////////////////////////////////////////

#define CONFIGURATION_CPP
#define BABY_IO
#include "Configuration.h"

class Configuration *config;

Configuration::Configuration()
{
  conf_file = "babyftpd.conf";
  parsed = false;

  general.connection_limit = INT_MAX; // ain't this a bit high?
  general.port = 21; // default port
  general.data_port = 20; // default outgoing data port
  general.no_daemon = false; // daemon mode or not?
  general.os = "UNIX"; // what OS is this?
  general.admin_email = ""; // admin email?
  general.lifebeat_tick = 300; // run the lifebeat() function every this
  char tmp[20];
  gethostname(tmp, 20);
  general.welcome_msg = "BabyFTPd ";
  general.welcome_msg += VERSION;
  general.welcome_msg += " Server [";
  general.welcome_msg += tmp;
  general.welcome_msg += "]";
  general.user = 0; // by default when run all as root
  general.group = 0;
  general.readonly = false;
  general.anonymous = A_DISABLE;
#ifdef USE_TLS
  general.tls_key = "/etc/babyftpd/tls_key.pem";
  general.tls_cert = "/etc/babyftpd/tls_cert.pem";
  general.tls_ca = "/etc/babyftpd/tls_ca.pem";
  general.tls_crl = "/etc/babyftpd/tls_crl.pem";
  general.dh_regenerate = 3600; // seconds between diffie-hellman regenerations
  general.verify_client = "/etc/babyftpd/verify_client.pl";

  general_user.encryption = ALLOW;
#endif
  logger.log_level = 3; // 3 is probably a sensible default?
  logger.xferlog = true; // by default all transfers are logged
  logger.xferlog_file = "/var/log/xferlog"; // here
  logger.thread_id = false; // we don't want thread ids all over the logs
  logger.time_format = "%b %e %H:%M:%S "; // time format

  // these options end up as default for all users, that is,
  //  those that don't occur in the conf file get these values
  general_user.id = "user_general";
  general_user.sim_data = INT_MAX;
  general_user.idle = 900; // idle timeout after how long?
  general_user.noftp = false; // can this user login with ftp or not?
  general_user.only_home = false; // false chroot in homedir?
  general_user.only_passive = false;
  general_user.user_limit = INT_MAX; // maximum control connections
}

Configuration::~Configuration()
{
}

bool Configuration::parse_config_file()
{
  vector<string> file_buff;
  if(!fill_buff(file_buff, conf_file))
    return(false);

  set_structs(get_variables(file_buff, "general", 1), GENERAL, &general);
  set_structs(get_variables(file_buff, "general", 2), USER, &general_user);
  set_structs(get_variables(file_buff, "log", 1), LOG, &logger);

  string user_names;
  num_users = count_users(file_buff, user_names) - 1; // first user is 0.

  class StringTokenizer *names = new StringTokenizer(user_names, ",", false);
  names->nextToken(); // not nice, but nicer.

  if(!users.empty()) // that is, on SIGHUP
    users.clear();

  users.reserve(num_users + 1);
  for(int i = 0; i < num_users; i++)
  {
    struct users_stru user_tmp = general_user;
    set_structs(get_variables(file_buff, names->nextToken().c_str(), 2),
        USER, &user_tmp);
    users.push_back(user_tmp);
  }

  parsed = true;
  delete names;
  return true;
}

bool Configuration::set_structs(string variables, enum sect section,
    void *this_struct)
{
  struct users_stru *usr_tmp = NULL;
  struct general_stru *gen_tmp = NULL;
  struct log_stru *log_tmp = NULL;
  vector<config_entry> this_entry;
  string for_log = "Configuration for ";

  switch(section)
  {
    case USER:
      {
        usr_tmp = (static_cast<struct users_stru*>(this_struct));

        this_entry.push_back(config_entry(
              "id", CONF_STRING, &usr_tmp->id));
        this_entry.push_back(config_entry(
              "sim_data", CONF_INT, &usr_tmp->sim_data));
        this_entry.push_back(config_entry(
              "idle", CONF_TIME, &usr_tmp->idle));
        this_entry.push_back(config_entry(
              "noftp", CONF_BOOL, &usr_tmp->noftp));
        this_entry.push_back(config_entry(
              "only_home", CONF_BOOL, &usr_tmp->only_home));
        this_entry.push_back(config_entry(
              "only_passive", CONF_BOOL, &usr_tmp->only_passive));
        this_entry.push_back(config_entry(
              "user_limit", CONF_INT, &usr_tmp->user_limit));
#ifdef USE_TLS
        this_entry.push_back(config_entry(
              "encryption", CONF_ENC, &usr_tmp->encryption));
#endif // USE_TLS
        for_log += "user section ";
      }
      break;

    case LOG:

      {
        log_tmp = (static_cast<struct log_stru*>(this_struct));

        this_entry.push_back(config_entry(
              "log_level", CONF_INT, &log_tmp->log_level));
        this_entry.push_back(config_entry(
              "xferlog", CONF_BOOL, &log_tmp->xferlog));
        this_entry.push_back(config_entry(
              "xferlog_file", CONF_LOGF, &log_tmp->xferlog_file));
        this_entry.push_back(config_entry(
              "authlog", CONF_LOGF, &log_tmp->authlog));
        this_entry.push_back(config_entry(
              "infolog", CONF_LOGF, &log_tmp->infolog));
        this_entry.push_back(config_entry(
              "debuglog", CONF_LOGF, &log_tmp->debuglog));
        this_entry.push_back(config_entry(
              "thread_id", CONF_BOOL, &log_tmp->thread_id));
        this_entry.push_back(config_entry(
              "time_format", CONF_TIMEF, &log_tmp->time_format));
        for_log += "log section ";
      }
      break;

    case GENERAL:
      {
        gen_tmp = (static_cast<struct general_stru*>(this_struct));

        this_entry.push_back(config_entry(
              "connection_limit", CONF_INT, &gen_tmp->connection_limit));
        this_entry.push_back(config_entry(
              "port", CONF_INT, &gen_tmp->port));
        this_entry.push_back(config_entry(
              "data_port", CONF_INT, &gen_tmp->data_port));
        this_entry.push_back(config_entry(
              "no_daemon", CONF_BOOL, &gen_tmp->no_daemon));
        this_entry.push_back(config_entry(
              "admin_email", CONF_STRING, &gen_tmp->admin_email));
        this_entry.push_back(config_entry(
              "welcome_msg", CONF_STRING, &gen_tmp->welcome_msg));
        this_entry.push_back(config_entry(
              "lifebeat_tick", CONF_TIME, &gen_tmp->lifebeat_tick));
        this_entry.push_back(config_entry(
              "user", CONF_USER, NULL));
        this_entry.push_back(config_entry(
              "pid_file", CONF_STRING, &gen_tmp->pid_file));
        this_entry.push_back(config_entry(
              "readonly", CONF_BOOL, &gen_tmp->readonly));
        this_entry.push_back(config_entry(
              "anonymous", CONF_ANON, &gen_tmp->anonymous));
#ifdef USE_TLS
        this_entry.push_back(config_entry(
              "tls_key", CONF_STRING, &gen_tmp->tls_key));
        this_entry.push_back(config_entry(
              "tls_cert", CONF_STRING, &gen_tmp->tls_cert));
        this_entry.push_back(config_entry(
              "tls_ca", CONF_STRING, &gen_tmp->tls_ca));
        this_entry.push_back(config_entry(
              "tls_crl", CONF_STRING, &gen_tmp->tls_crl));
        this_entry.push_back(config_entry(
              "dh_regenerate", CONF_TIME, &gen_tmp->dh_regenerate));
        this_entry.push_back(config_entry(
              "verify_client", CONF_STRING, &gen_tmp->verify_client));
#endif // USE_TLS
        for_log += "general section ";
      }
      break;
  }
  class StringTokenizer *var = new StringTokenizer(variables, "\n", false);
  string id = var->nextToken();
  if(section == USER)
  {
    *(string*)this_entry[0].variabel = id;
    for_log += "for " + id + " ";
  }

  while(var->hasMoreTokens())
  {
    string log_me, str = var->nextToken();
    class StringTokenizer *st = new StringTokenizer(str, " ", false);
    string op = st->nextToken();
    string arg = st->restTokens();
    vector<config_entry>::iterator this_it = this_entry.begin();

    for(; this_it != this_entry.end(); this_it++)
      if(op == (*this_it).name)
        break;

    if(this_it == this_entry.end())
      log_me = "Parameter " + op + " doesn't exist, check your spelling.";
    else
    {
      switch((*this_it).type)
      {
        case CONF_TIME:
        case CONF_ANON:
#ifdef USE_TLS
        case CONF_ENC:
          if((*this_it).type == CONF_ENC)
          {
            if(arg == "allow")
              arg = util.itos(ALLOW);
            else if(arg == "deny")
              arg = util.itos(DENY);
            else if(arg == "force")
              arg = util.itos(FORCE);
            else if(arg == "force_cert")
              arg = util.itos(FORCE_CERT);
            else
              log_me = "only allow, force, foce_cert and deny are legal "
                "parameters to encryption, " + arg + " isn't one of them.";
          }
          else
#endif // USE_TLS
            if((*this_it).type == CONF_TIME)
            {
              int multi = 1;
              string::iterator st_it = arg.begin();
              for(; st_it != arg.end(); st_it++)
              {
                if(isdigit(*st_it) || isspace(*st_it))
                  continue;
                else
                  break;
              }
              if(st_it != arg.end())
              {
                switch(*st_it)
                {
                  case 's':
                    multi = 1;
                    break;
                  case 'm':
                    multi = 60;
                    break;
                  case 'h':
                    multi = 3600;
                    break;
                  case 'd':
                    multi = 86400;
                    break;
                  default:
                    log_me = "I don't understand your time unit \'";
                    log_me += *st_it;
                    log_me += "\' to paramater " + op + " I will ignore it.";
                }
                if(!log_me.empty())
                  break;

                arg.erase(st_it, arg.end());
                int temp_num = atoi(arg.c_str()) * multi;
                arg = util.itos(temp_num);
              }
            }
            else if((*this_it).type == CONF_ANON)
            {
              if(arg == "enable")
                arg = util.itos(A_ENABLE);
              else if(arg == "disable")
                arg = util.itos(A_DISABLE);
              else if(arg == "only")
                arg = util.itos(A_ONLY);
              else
                log_me = "only enable, disable and only are legal "
                  "parameters to anonymous, " + arg + " isn't one of them.";
            }
        case CONF_INT:
          {
            int i_value = 0;
            bool error = false;
            for(string::iterator it = arg.begin(); it != arg.end(); it++)
              if(!isdigit(*it))
              {
                error = true;
                break;
              }
            if(!error)
              i_value = atoi(arg.c_str());
            else
              if(log_me.empty())
                log_me = "Non int value supplied to " + op +
                  " value supplied was " + arg + ".";

            if(log_me.empty())
            {
              *(static_cast<int*>((*this_it).variabel)) = i_value;
              for_log += (*this_it).name + "=" + util.itos(i_value) + "; ";
            }
          }
          break;

        case CONF_BOOL:
          {
            bool b_value = true;
            if(arg == "true" || arg == "t" || arg == "1" || arg == "y")
              b_value = true;
            else if(arg == "false" || arg == "f" || arg == "0" || arg == "n")
              b_value = false;
            else
              log_me = arg + " not recognised as a boolean value supplied to "
                + op + ".";

            if(log_me.empty())
            {
              *(static_cast<bool*>((*this_it).variabel)) = b_value;
              for_log += (*this_it).name + (b_value ? "=true; " : "=false; ");
            }
          }
          break;

        case CONF_USER:
          {
            struct passwd pwd, *pwd_ptr;
            struct group grp, *grp_ptr;
            char pwdata[PWBUF_SIZE], grpdata[PWBUF_SIZE];
            int delim = arg.find_first_of(':');
            getpwnam_r(arg.substr(0, delim).c_str(),
                &pwd, pwdata, sizeof(pwdata), &pwd_ptr);
            getgrnam_r(arg.substr(delim + 1).c_str(),
                &grp, grpdata, sizeof(grpdata), &grp_ptr);
            if(pwd_ptr == NULL)
            {
              log_me = "Tried to use nonexistent user " + arg.substr(0,
                  delim) + " to run the daemon under. using root instead.";
              pwd.pw_uid = 0;
              grp.gr_gid = 0;
            }
            else if(grp_ptr == NULL)
            {
              log_me = "Tried to use nonexistent group " + arg.substr(delim +1)
                + " to run the daemon under. using root instead.";
              grp.gr_gid = 0;
            }
            general.user = pwd.pw_uid;
            general.group = grp.gr_gid;
            for_log += (*this_it).name + "=" +
              util.itos(pwd.pw_uid) + ":" + util.itos(grp.gr_gid) + "; ";
          }
          break;

        case CONF_LOGF:
          if(arg == "syslog")
            arg.clear();
        case CONF_TIMEF:
          if((*this_it).type == CONF_TIMEF)
            arg += " ";
        case CONF_STRING:
          *(static_cast<string*>((*this_it).variabel)) = arg;
          for_log += (*this_it).name + "=" + arg + "; ";
          break;

        default:
          log_me = "Parameter " + op + " doesn't exist, check your spelling.";
          break;
      }
    }
    delete st;
    if(!log_me.empty())
      logging->log_this(2, TYPE_INFO, log_me);
  }

  if(this->general.anonymous >= A_ENABLE && this->general.user == 0)
  {
    this->general.anonymous = A_DISABLE;
    logging->log_this(2, TYPE_INFO,
        "you enabled anonymous logins but you don't have a configured user."
        " anonymous logins disabled.");
  }

  logging->log_this(3, TYPE_INFO, for_log);
  delete var;
  return true;
}

struct users_stru &Configuration::get_user(string name)
{
  for(vector<users_stru>::iterator use_it = users.begin();
      use_it != users.end(); use_it++)
    if((*use_it).id == name && name != "")
      return(*use_it);
  return general_user;
}

int Configuration::count_users(vector<string> buff, string &users)
{
  int ant = 0;
  int deep = 1, curr_deep = 0;
  string name = "user", section = "";
  for(vector<string>::iterator buff_it = buff.begin(); buff_it != buff.end();
      buff_it++)
  {
    if(*buff_it == "{")
    {
      if(curr_deep == deep && section == name)
      {
        ant++;
        users += *(buff_it - 1) + ",";
      }
      curr_deep++;
      if(curr_deep == deep)
        section = *(buff_it - 1);
    }
    else if(*buff_it == "}")
    {
      if(curr_deep == deep && section == name)
        return ant;
      curr_deep--;
    }
  }
  return ant;
}


string Configuration::get_variables(vector<string> &buff, const char *key,
    int deep)
{
  int curr_deep = 0;
  string section, str;
  for(vector<string>::iterator buff_it = buff.begin(); buff_it != buff.end();
      buff_it++)
  {
    if((*buff_it).empty())
      continue;
    size_t comment = (*buff_it).find_first_of('#');
    if(comment != string::npos)
      (*buff_it).erase(comment);
    if(*buff_it == "{")
    {
      section = *(buff_it - 1);
      curr_deep++;
      if(curr_deep == deep)
        str = section + "\n";
    }
    else if(*buff_it == "}")
    {
      if(curr_deep == deep && section == key)
        break;
      curr_deep--;
    }
    else if(curr_deep == deep && section == key)
      str += *buff_it + "\n";
  }
  return(str);
}

bool Configuration::fill_buff(vector<string> &buff, string filename)
{
  ifstream fin;
  fin.open(filename.c_str(), ifstream::in);

  while(fin.good())
  {
    string buffer;
    getline(fin, buffer);
    util.clean_string(buffer);
    if(!fin.good() || buffer.empty())
      continue;
    buff.push_back(buffer);
  }

  fin.close();
  if(buff.empty())
    return(false);
  else
    return(true);
}


bool Configuration::parse_command_line(int argc, char **argv)
{
  int opt = 0;
  const char *cmdopts = "hc:nVl:p:P:";

  // struct for the long options... neat, huh?
  struct option longopts[] =
  {
    { "help",      0, 0, 'h' },
    { "conf-file", 1, 0, 'c' },
    { "no-daemon", 0, 0, 'd' },
    { "version",   0, 0, 'V' },
    { "log-level", 1, 0, 'l' },
    { "port",      1, 0, 'p' },
    { "pid_file",  1, 0, 'P' },
    { 0,           0, 0, 0   }
  };

  while((opt = getopt_long(argc, argv, cmdopts, longopts, NULL)) != -1)
  {
    switch(opt)
    {
      case 'h':
        cout << "Usage: " << argv[0] << " [OPTION]...\n\n"
          << "  -h, --help            display this help and exit\n"
          << "  -c, --conf-file=FILE  supply another conf file instead "
          << "of the default\n"
          << "  -n, --no-daemon       do not fork into the background\n"
          << "  -V, --version         display version information and exit\n"
          << "  -l, --log-level=LEVEL set the log level, 1-10, default is 5\n"
          << "  -p, --port=PORT       specify an alternate port\n"
          << "  -P, --pid_file=FILE   a file to write our pid in\n\n"
          << "Please, report bugs, either directly to the developers\n"
          << "or via sourceforge (http://www.sf.net/projects/babyftpd).\n";

          exit_baby(0, EXIT_EARLY);

      case 'c':
        config->conf_file = optarg;
        break;

      case 'n':
        config->general.no_daemon = true;
        break;

      case 'V':
        cout << "BabyFTPd version " << VERSION << "\nBuilt on " << __DATE__
          << " " << " " << __TIME__ << endl;
        exit_baby(0, EXIT_EARLY);

      case 'l':
        // Log level code...
        if(atoi(optarg) > 10)
          logger.log_level = 10;
        else if(atoi(optarg) < 0)
          logger.log_level = 0;
        else
          logger.log_level = atoi(optarg);
        break;

      case 'p':
        // alternate port...
        config->general.port = atoi(optarg);
        break;

      case 'P':
        // pid_file
        config->general.pid_file = optarg;
    }
  }
  return true;
}

