/////////////////////////////////////////////////////////////////////
//  $Id: Configuration.h,v 1.55 2005/08/23 19:34:02 klas Exp $
//
//  Usage: Configure class, reading the config file and holding
//           configurish variables.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file                
//    LICENSE for complete information.                                         
/////////////////////////////////////////////////////////////////////


#ifndef CONFIGURATION_H
#define CONFIGURATION_H

#ifdef CONFIGURATION_CPP
#include "generic.h"
#include "StringTokenizer.h"
#include "Utilities.h"
#include "Log.h"
extern void exit_baby(int, enum exits);
const char *VERSION = "CVS: "__DATE__; // remember to change on release.
#endif

struct log_stru
{
  int log_level;
  bool xferlog;
  string xferlog_file;
  string authlog;
  string infolog;
  string debuglog;
  bool thread_id;
  string time_format;
};

struct users_stru
{
  string id;
  int sim_data;
  int idle;
  bool noftp;
  bool only_home;
  bool only_passive;
  int user_limit;
#ifdef USE_TLS
  int encryption;
#endif // USE_TLS
};

struct general_stru
{
  int connection_limit;
  int port;
  int data_port;
  bool no_daemon;
  string os;
  string admin_email;
  string welcome_msg;
  int lifebeat_tick;
  uid_t user;
  gid_t group;
  string pid_file;
  bool readonly;
  int anonymous;
#ifdef USE_TLS  
  string tls_key;
  string tls_cert;
  string tls_ca;
  string tls_crl;
  int dh_regenerate;
  string verify_client;
#endif
};

enum sect { USER, LOG, GENERAL };

class Configuration
{
  private:
    vector<users_stru> users;
    int num_users;
    bool parsed;
    int count_users(vector<string>, string&);
    bool fill_buff(vector<string>&, string);
    string get_variables(vector<string>&, const char*, int);
    bool set_structs(string, enum sect, void*);
  public:
    bool parse_config_file();
    bool parse_command_line(int argc, char **argv);
    string conf_file;
    struct log_stru logger;
    struct general_stru general;
    struct users_stru general_user;
    Configuration(void);
    struct users_stru &get_user(string);
    ~Configuration(void);
};

extern class Configuration *config;

enum conf
{
  CONF_INT, CONF_BOOL, CONF_STRING, CONF_LOGF, CONF_ENC,
  CONF_USER, CONF_TIMEF, CONF_TIME, CONF_ANON
};

struct config_entry
{
  string name;
  enum conf type;
  void *variabel;
  config_entry(const string t_name, const enum conf t_type, void *t_vari)
  {
    name = t_name;
    type = t_type;
    variabel = t_vari;
  }
};

typedef struct config_entry config_entry;

// constants for the anonymous variable.
enum anon { A_DISABLE, A_ENABLE, A_ONLY };

#ifdef USE_TLS
enum encr { ALLOW, FORCE, DENY, FORCE_CERT };
#endif // USE_TLS
#endif
