//////////////////////////////////////////////////////////////////////////
// $Id: Connection.h,v 1.19 2005/09/01 20:48:22 klas Exp $
//
// Usage: Connection class. is inherited to user and data.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
//////////////////////////////////////////////////////////////////////////

#ifndef CONNECTION_H
#define CONNECTION_H

#ifdef CONNECTION_CPP
#include "generic.h"
#include "Utilities.h"
#include "Configuration.h"
#include "Log.h"
#endif // CONNECTION_CPP

// for settings()
enum bits { B_SET, B_RESET, B_FLIP };

enum sets { // for mlsx
  FACT_TYPE, FACT_SIZE, FACT_MODIFY, FACT_UNIQUE, FACT_U_MODE, FACT_U_UID,
  FACT_U_GID, FACT_U_LINK, FACT_PERM,
  DATA_ACCT, BINARY, COMPRESSED, PASSIVE, ANONYMOUS, DATA_OPEN, EPSV_ONLY,
  RETR_TRANS, PAM_DONE, FAILURE
#ifdef USE_TLS
    , TLS_ENABLED, TLS_INITED, TLS_PBSZ, TLS_PROT_D
#endif // USE_TLS
};

const int LAST_FACT = 8;

// keep an eye on these.. different depending on TLS or not...
#ifdef USE_TLS
const int TOTAL_U_BIT = 23;
#else
const int TOTAL_U_BIT = 19;
#endif // USE_TLS

// macro that builds get and set functions for us.
#define MEMBER(func, type)  private: type _##func; \
  public: type func(void){ \
            BABY_RDLOCK(this->lock); \
            type t = _##func; \
            BABY_UNLOCK(this->lock); \
            return t;} \
  public: void func(type var){ \
            BABY_WRLOCK(this->lock); \
            this->_##func = var; \
            BABY_UNLOCK(this->lock); \
          }

class Connection
{
  public:
    pthread_rwlock_t lock;
    ssize_t send(const char*, size_t);
    ssize_t receive(char*, size_t);
    Connection &operator<<(long long int num);
    Connection &operator<<(string message);
#ifdef USE_TLS
    bool engage_tls();
    bool disengage_tls();
    bool verify_cert();
#endif // USE_TLS
    bitset<TOTAL_U_BIT> settings() const
    {
      return _settings;
    }
    bool settings(int bit)
    {
      assert(bit < TOTAL_U_BIT);
      BABY_RDLOCK(this->lock);
      bool b = this->_settings.test(bit);
      BABY_UNLOCK(this->lock);
      return(b);
    }
    void settings(int bit, int way)
    {
      BABY_WRLOCK(this->lock);
      if(way == B_SET)
        this->_settings.set(bit);
      else if(way == B_RESET)
        this->_settings.reset(bit);
      else if(way == B_FLIP)
        this->_settings.flip(bit);
      BABY_UNLOCK(this->lock);
    }
  protected:
    Connection();
    ~Connection();
    bitset<TOTAL_U_BIT> _settings;
    MEMBER(thread_id, pthread_t);
    MEMBER(host_name, string);
    MEMBER(socket_id, int);
    MEMBER(addr, struct sockaddr_in);
    MEMBER(packet_size, size_t);
    MEMBER(timeout, struct timeval);
    MEMBER(user_name, string);
    MEMBER(transfered, unsigned long long);
  private:
#ifdef USE_TLS
    gnutls_session session;
#endif // USE_TLS
};

const int READ_SIZE = 16384;

extern vector<pthread_t> cleanup_list;
extern sem_t cleanup_list_lock;
#ifdef USE_TLS
extern gnutls_certificate_credentials x509_cred;
extern int tls_db_store(void*, gnutls_datum, gnutls_datum);
extern gnutls_datum tls_db_fetch(void*, gnutls_datum);
extern int tls_db_delete(void*, gnutls_datum);
#endif // USE_TLS

#endif // CONNECTION_H
