//////////////////////////////////////////////////////////////////////////
// $Id: Data.cpp,v 1.13 2005/08/23 19:32:31 klas Exp $
//
// Usage: Data class. inherited from Connection, everything pertaining to
//        data connection goes here.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
//////////////////////////////////////////////////////////////////////////

#define DATA_CPP
#define NETINCLS
#define BABY_CORE
#include "Data.h"

Data::Data(User *user, int file_id, string file_name, int trans_type)
  : Connection()
{
  this->thread_id(0); // data_wrapper sets this to it's own.
  this->_file_id = file_id;
  this->_file_name = file_name;
  this->user = user;
  this->_settings = user->settings();
  this->addr(user->data_addr());
  struct timeval tv = {1, 0};
  this->timeout(tv);
  if(trans_type == 1)
    this->settings(RETR_TRANS, B_SET);

  this->settings(DATA_OPEN, B_SET);
#ifdef USE_TLS
  this->settings(TLS_INITED, B_RESET);
  this->settings(TLS_ENABLED, B_RESET);
#endif // USE_TLS

  if(this->settings(PASSIVE))
    this->open_passive();
  else // ACTIVE
    this->open_active();

  // if the data connection comes from another host than the
  // ctrl connection we drop it.. unless the user is tls protected.
  if(this->addr().sin_addr.s_addr != user->addr().sin_addr.s_addr)
  {
#ifdef USE_TLS
    if(!user->settings(TLS_ENABLED))
    {
#endif // USE_TLS
      close(this->socket_id());
      this->settings(DATA_OPEN, B_RESET);
    }
#ifdef USE_TLS
  }
#endif //USE_TLS
}

void Data::open_passive()
{
  socklen_t len = sizeof(this->addr());
  struct sockaddr_in temp_addr = this->addr();

  fd_set acc_set;
  FD_ZERO(&acc_set);
  FD_SET(user->data_id(), &acc_set);
  struct timeval tv = this->timeout();

  if(select(user->data_id() + 1, &acc_set, NULL, NULL, &tv) > 0)
  {
    int ret = accept(user->data_id(),
        reinterpret_cast<struct sockaddr*>(&temp_addr), &len);
    this->socket_id(ret);
    this->addr(temp_addr);
  }
  else
    this->settings(DATA_OPEN, B_RESET);

  close(user->data_id());
}

void Data::open_active()
{
  this->socket_id(socket(AF_INET, SOCK_STREAM, IPPROTO_TCP));

  int on = 1;
  setsockopt(this->socket_id(), SOL_SOCKET, SO_REUSEADDR,
      (static_cast<void*>(&on)), sizeof(on));
  struct sockaddr_in my_addr;
  my_addr.sin_port = htons(config->general.data_port);
  my_addr.sin_addr.s_addr = htonl(INADDR_ANY);

  uid_t old_uid = 0;
  // if dataport is below 1024 we need root.
  if(config->general.data_port <= 1024)
  {
    old_uid = geteuid();
    seteuid(0);
  }
  int ret = bind(this->socket_id(), reinterpret_cast<struct sockaddr*>
      (&my_addr), sizeof(my_addr));
  if(config->general.data_port <= 1024)
    seteuid(old_uid);

  if(ret != 0)
    logging->log_this(3, TYPE_INFO,
        "Failed to bind to data port " + util.itos(config->general.data_port)
        + ", falling back to arbitrary port.");

  struct sockaddr_in tmp_addr = this->addr();

  // go into nonblocking
  long sock_arg = fcntl(this->socket_id(), F_GETFL, NULL);
  sock_arg |= O_NONBLOCK;
  fcntl(this->socket_id(), F_SETFL, sock_arg);

  int sock_err = connect(this->socket_id(), reinterpret_cast<struct sockaddr*>
        (&tmp_addr), sizeof(tmp_addr));
  this->addr(tmp_addr);

  if(sock_err < 0)
    if(errno == EINPROGRESS)
    {
      struct timeval tv = this->timeout();
      fd_set con_set;
      FD_ZERO(&con_set);
      FD_SET(this->socket_id(), &con_set);
      if(select(this->socket_id() + 1, NULL, &con_set, NULL, &tv) < 0)
        this->settings(DATA_OPEN, B_RESET);
    }
    else
      this->settings(DATA_OPEN, B_RESET);

  // leave non blocking
  long sock_arg2 = fcntl(this->socket_id(), F_GETFL, NULL);
  sock_arg2 &= (~O_NONBLOCK);
  fcntl(this->socket_id(), F_SETFL, sock_arg2);
}

Data::~Data()
{
  if(this->_file_id != -1)
    close(this->_file_id);

  this->user->transfered(
      this->user->transfered() + this->transfered());

  if(this->thread_id())
  {
    BABY_WRLOCK(user->lock);
    user->data_vec.erase(user->get_data_it(this->thread_id()));
    BABY_UNLOCK(user->lock);
    logging->log_this(4, TYPE_INFO, "data thread " +
        util.itos(this->thread_id()) +
        " exiting. Remaining data connections for " + user->user_name() +
        " is: " + util.itos(user->data_vec.size()));
  }
}

void kill_data(void *data)
{
  delete static_cast<Data*>(data);
}
