/////////////////////////////////////////////////////////////////////
// $Id: Handler.cpp,v 1.162 2005/07/23 19:57:17 klas Exp $
//
// Usage: the main loop for connected clients and interpreter for
//        ftp commands.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define HANDLER_CPP
#define NETINCLS
#define BABY_CORE
#include "Handler.h"
#include "commands.hlp"

class Handler *handler;

Handler::Handler()
{
  // first argument is number of arguments to the command, they should
  //  be self explaining except the negative ones. -1 means any number,
  //  -2 means anything but 0.
  // An empty hlp field means no help will be displayed about the command
  cmd_info["SYST"] = command_info(0, &Commands::sim1, COM_SYST, HLP_SYST);
  cmd_info["HELP"] = command_info(-1, &Commands::help, BEFORE_USER, HLP_HELP);
  cmd_info["RETR"] = command_info(-2, &Commands::retr, 0, HLP_RETR);
  // these commands are disabled if the conf variable readonly in general
  //  sector is enabled.
  if(!config->general.readonly)
  {
    cmd_info["STOR"] = command_info(-2, &Commands::stor, COM_STOR, HLP_STOR);
    cmd_info["APPE"] = command_info(-2, &Commands::stor, COM_APPE, HLP_APPE);
    cmd_info["DELE"] = command_info(-2, &Commands::dele, 0, HLP_DELE);
    cmd_info["RNFR"] = command_info(-2, &Commands::rname, COM_FROM, HLP_RNFR);
    cmd_info["RNTO"] = command_info(-2, &Commands::rname, COM_TO, HLP_RNTO);
    cmd_info["MKD"] = command_info(-2, &Commands::mkd, 0, HLP_MKD);
    cmd_info["RMD"] = command_info(-2, &Commands::rmd, 0, HLP_RMD);
    cmd_info["XMKD"] = command_info(-2, &Commands::mkd, 0, HLP_MKD);
    cmd_info["XRMD"] = command_info(-2, &Commands::rmd, 0, HLP_RMD);
  }
  cmd_info["PORT"] = command_info(1, &Commands::port, COM_PORT, HLP_PORT);
  cmd_info["EPRT"] = command_info(1, &Commands::port, COM_EPRT, HLP_EPRT);
  cmd_info["PASV"] = command_info(0, &Commands::pasv, COM_PASV, HLP_PASV);
  cmd_info["EPSV"] = command_info(-1, &Commands::pasv, COM_EPSV, HLP_EPSV);
  cmd_info["PWD"] = command_info(0, &Commands::sim1, COM_PWD, HLP_PWD);
  cmd_info["CWD"] = command_info(-2, &Commands::cwd,  COM_CWD, HLP_CWD);
  cmd_info["CDUP"] = command_info(0, &Commands::cwd,  COM_CDUP, HLP_CDUP);
  cmd_info["LIST"] = command_info(-1, &Commands::list, COM_LONG, HLP_LIST);
  cmd_info["NLST"] = command_info(-1, &Commands::list, COM_SHORT, HLP_NLST);
  cmd_info["TYPE"] = command_info(-2, &Commands::sim1, COM_TYPE, HLP_TYPE);
  cmd_info["MODE"] = command_info(1, &Commands::sim1, COM_MODE, HLP_MODE);
  cmd_info["STRU"] = command_info(1, &Commands::sim1, COM_STRU, HLP_STRU);
  cmd_info["SIZE"] = command_info(-2, &Commands::sim2, COM_SIZE, HLP_SIZE);
  cmd_info["USER"] = command_info(1, &Commands::user, BEFORE_USER, HLP_USER);
  cmd_info["PASS"] = command_info(-2, &Commands::pass, BEFORE_USER, HLP_PASS);
  cmd_info["QUIT"] = command_info(-1, &Commands::quit, BEFORE_USER, HLP_QUIT);
  cmd_info["NOOP"] = command_info(0, &Commands::sim1, COM_NOOP, HLP_NOOP);
  cmd_info["MDTM"] = command_info(-2, &Commands::sim2, COM_MDTM, HLP_MDTM);
  cmd_info["FEAT"] =
    command_info(0, &Commands::sim1, COM_FEAT | BEFORE_USER, HLP_FEAT);
  cmd_info["MLST"] = command_info(-2, &Commands::sim2, COM_MLST, HLP_MLST);
  cmd_info["MLSD"] = command_info(-1, &Commands::list, COM_MLSD, HLP_MLSD);
  cmd_info["OPTS"] = command_info(-1, &Commands::opts, 0, HLP_OPTS);
  cmd_info["REST"] = command_info(1, &Commands::rest, 0, HLP_REST);
  cmd_info["XCWD"] = command_info(-2, &Commands::cwd, COM_CWD, HLP_CWD);
  cmd_info["XPWD"] = command_info(0, &Commands::sim1, COM_PWD, HLP_PWD);
  cmd_info["XCUP"] = command_info(0, &Commands::sim1, COM_CDUP, HLP_CDUP);
#ifdef USE_TLS
  cmd_info["AUTH"] = command_info(1, &Commands::auth, BEFORE_USER, HLP_AUTH);
  cmd_info["PBSZ"] =
    command_info(1, &Commands::sim1, COM_PBSZ | BEFORE_USER, HLP_PBSZ);
  cmd_info["PROT"] =
    command_info(1, &Commands::sim1, COM_PROT | BEFORE_USER, HLP_PROT);
  cmd_info["CCC"] = command_info(0, &Commands::sim1, COM_CCC, HLP_CCC);
  // these commands are required to be understood by rfc2228, even if
  // we don't use them.
  cmd_info["CONF"] =
    command_info(-1, &Commands::sim1, COM_CONF | BEFORE_USER, "");
  cmd_info["MIC"] =
    command_info(-1, &Commands::sim1, COM_MIC | BEFORE_USER, "");
  cmd_info["ENC"] =
    command_info(-1, &Commands::sim1, COM_ENC | BEFORE_USER, "");
#endif // USE_TLS
  cmd_info["ABOR"] = command_info(0, &Commands::abor, 0, HLP_ABOR);
}

Handler::~Handler()
{
  cmd_info.clear();
}

class User *Handler::initializer(int socket_id)
{
  struct sockaddr_in addr;
  socklen_t len = sizeof(sockaddr_in);

  getpeername(socket_id, reinterpret_cast<struct sockaddr*>(&addr), &len);
  class User *user = new User(socket_id, addr);

  string log = "new connection from: ";
  log += inet_ntoa(addr.sin_addr);
  logging->log_this(3, TYPE_INFO, log);

  pthread_rwlock_wrlock(&user_list_lock);
  user_list.push_back(user);
  pthread_rwlock_unlock(&user_list_lock);

  string message;
  pthread_rwlock_rdlock(&user_list_lock);
  if(static_cast<int>(user_list.size()) > config->general.connection_limit)
  {
    // reached maximum number of connections...
    message =
      "421 Maximum number of connections reached, please try again later.";
    user->settings(FAILURE, B_SET);
  }
  else
  {
    // dns lookup should go here.. find a nice way of doing it.
    message = "220 " + config->general.welcome_msg;
  }
  pthread_rwlock_unlock(&user_list_lock);
  *user << message << "\r\n";

  return(user);
}

bool Handler::init(User *user)
{
  while(true)
  {
    string command;
    if(user->settings(FAILURE))
      break;
    *user >> command;
    util.clean_string(command);
    if(!command.empty())
      this->interpreter(user, command);
  }
  return(true); // we're returning to wrapper for exit.
}

bool Handler::interpreter(User *user, string command)
{
  string argument, message;
  class StringTokenizer *st = NULL;

  st = new StringTokenizer(command, " ", false);
  int words = st->countTokens() - 1;
  command = st->nextToken();

  transform(command.begin(), command.end(), command.begin(),
      (int(*)(int))toupper);

  if(command == "SITE")
  {
    st->nextToken();
    command += " " + st->nextToken();
    transform(command.begin() + 4, command.end(), command.begin() + 4,
        (int(*)(int))toupper);
    words--;
  }
  argument = st->restTokens();
  delete(st);

  if(command != "PASS")
    logging->log_this(5, TYPE_DEBUG, command + "-" + argument);
  else
    logging->log_this(5, TYPE_DEBUG, "PASS-**********");

  hash_map<const char*, struct command_info, hash<const char*>,
    Utilities::eqstr>::iterator cmd_it = cmd_info.find(command.c_str());

  if(cmd_it == cmd_info.end())
    message = "500 '" + command + "' not understood.";
  else
  {
    if(user->logged_in() < LOGGED_IN &&
        !util.is_set((*(cmd_it)).second.extra_flag,BEFORE_USER))
      message = "530 Login with 'USER' and 'PASS' first.";
    else
    {
      if((!user->temp_buf().empty()) && command != "RNTO")
        user->temp_buf(""); // we don't want this hanging around forever.

      if(words == 0 && (((*(cmd_it)).second.ant_arg == -2) ||
            (words < (*(cmd_it)).second.ant_arg)))
        message = "501 '" + command + "': requires an argument.";
      else if((*(cmd_it)).second.ant_arg >= 0 &&
          words > (*(cmd_it)).second.ant_arg)
        message = "501 '" + command + "': to many arguments to command.";
      else
        ((comm.*(*(cmd_it)).second.pf)
         (user, argument, (*(cmd_it)).second.extra_flag));
    }
  }
  if(!message.empty())
    *user << message << "\r\n";
  return(true);
}
