/////////////////////////////////////////////////////////////////////
// $Id: StringTokenizer.cpp,v 1.17 2005/04/12 21:18:06 klas Exp $
//
// Usage: Used to chop up a string in "tokens"
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define STRINGTOKENIZER_CPP
#include "StringTokenizer.h"

/*
This class is made to handel strings. The StringTokenizer chops a string 
up into tokens using delimeters defined in a string passed ass an argument 
to the constructor.

The idea of writing this class comes from a similar class in java. but this 
is more functional and works in a better way.

The construktor takes three arguments, two strings and one bool. 
The first string is the string you want to make tokens of.
The secound string is a string containing all the characters that are 
going to cut the string.
The bool determins if the token chars should come with the tokens or not.
If the bool is set to false, the chars used to chop up the string is taken 
away.
*/

//Original constructor
StringTokenizer::StringTokenizer(string str, string delim = " \t\n\r\f",
    bool returnDelims = false)
{
  this->text = str;
  this->currentPosition = text.begin();
  this->delimiters = delim;
  this->retDelims = returnDelims;
}

//Private function used to get the position of the token.
string::const_iterator
StringTokenizer::scanToken(string::const_iterator startPos) const
{
  while (startPos != this->text.end())
  {
    if (isDelimiter(*startPos))
    {
      if(retDelims)
        startPos++;
      break;
    }
    startPos++;
  }
  return startPos;
}

//Private function that returns a position so that the delimiters is skipped.
string::const_iterator
StringTokenizer::skipDelimiters(string::const_iterator startPos) const
{
  while (!retDelims && startPos != this->text.end())
  {
    if (!isDelimiter(*startPos) || (delimiters.find(*startPos) == string::npos))
      break;
    startPos++;
  }
  return startPos;
}

//Public function that checks if the string got more tokens.
bool StringTokenizer::hasMoreTokens() const
{
  return((skipDelimiters(this->currentPosition) == this->text.end())
      ? false : true);
}

//Public function that returns the next token.
string StringTokenizer::nextToken()
{
  this->currentPosition = skipDelimiters(this->currentPosition);

  if (this->currentPosition == this->text.end())
    logging->log_this(2, TYPE_INFO, "NoSuchElementException.");

  string::const_iterator start = this->currentPosition;
  this->currentPosition = scanToken(this->currentPosition);
  return string(start, this->currentPosition);
}

//Public functoin that retutns the number of tokens the string is containing.
int StringTokenizer::countTokens() const
{
  int count = 0;
  string::const_iterator currpos = this->currentPosition;
  while (currpos != this->text.end()) {
    currpos = skipDelimiters(currpos);
    if (currpos == this->text.end())
      break;
    currpos = scanToken(currpos);
    count++;
  }
  return count;
}

//Public function that returns all the tokens after the last used.
string StringTokenizer::restTokens(void) const
{
  return(string((isDelimiter(*this->currentPosition) ? this->currentPosition
          + 1 :this->currentPosition), this->text.end()));
}
