/////////////////////////////////////////////////////////////////////
// $Id: User.cpp,v 1.98 2005/07/23 19:57:17 klas Exp $
//
// Usage: Holds information about connected users.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define USER_CPP
#define NETINCLS
#define BABY_CORE
#include "User.h"

User::User(int socket_id, struct sockaddr_in &addr) : Connection()
{
  this->_root_dir = "/";
  this->_failed_logins = 0;
  this->_logged_in = NOT_LOGGED_IN;
  this->_data_id = 0;
  this->_rest_pos = 0;
  this->addr(addr);
  this->socket_id(socket_id);
  this->thread_id(pthread_self());
  this->host_name(inet_ntoa(this->addr().sin_addr));
  struct timeval tmp = {config->general_user.idle, 0};
  this->timeout(tmp);

  // setting the bitflags on mlst_facts... not to clean..
  for(int i = 0; i <= LAST_FACT; i++)
    this->settings(i, B_SET);
}

User::~User()
{
  if(this->_data_id)
    close(this->_data_id);

  if(this->settings(PAM_DONE))
  {
    sem_wait(&comm.pam_lock);
    pam_close_session(this->_pamh, PAM_SILENT);
    pam_end(this->_pamh, PAM_SUCCESS);
    sem_post(&comm.pam_lock);
  }

  // killing off any ongoing data threads
  for(vector<Data*>::iterator data_it = data_vec.begin();
      data_it != data_vec.end(); data_it++)
    if(*data_it)
      pthread_cancel((*data_it)->thread_id());

  while(data_vec.size())
  {  }

  pthread_rwlock_wrlock(&user_list_lock);
  user_list.erase(get_user_it(this->thread_id()));
  logging->log_this(4, TYPE_INFO, "user thread " +
      util.itos(this->thread_id()) + " exiting, connected users:" +
      util.itos(user_list.size()));
  pthread_rwlock_unlock(&user_list_lock);
}

vector<User*>::iterator get_user_it(pthread_t thread_id)
{
  vector<User*>::iterator user_it;

  BABY_RDLOCK(user_list_lock);
  for(user_it = user_list.begin(); user_it != user_list.end(); user_it++)
    if(pthread_equal((*user_it)->thread_id(), thread_id))
      break;
  BABY_UNLOCK(user_list_lock);

  return(user_it);
}


vector<Data*>::iterator User::get_data_it(pthread_t thread_id)
{
  vector<Data*>::iterator data_it;

  BABY_RDLOCK(this->lock);
  for(data_it = this->data_vec.begin(); data_it != this->data_vec.end();
      data_it++)
    if(pthread_equal((*data_it)->thread_id(), thread_id))
      break;
  BABY_UNLOCK(this->lock);

  return(data_it);
}

bool User::data_vec_full()
{
  if(data_vec.size() >=
      static_cast<size_t>(config->get_user(this->user_name()).sim_data))
    return(true);
  return(false);
}

bool User::login(void)
{
  char pwdata[PWBUF_SIZE];
  struct passwd pwd, *pwd_ptr;
  this->logged_in(LOGGED_IN);
  getpwnam_r(this->user_name().c_str(), &pwd, pwdata, sizeof(pwdata),
      &pwd_ptr);
  uid_t uid = pwd.pw_uid;
  gid_t gid = pwd.pw_gid;
  if(config->get_user(this->user_name()).only_home)
    this->root_dir(pwd.pw_dir);
  this->current_dir(pwd.pw_dir);
  seteuid(0);
  setgid(gid);
  seteuid(uid); // dropping permissions...
  struct timeval tmp = {config->get_user(this->user_name()).idle, 0};
  this->timeout(tmp);
  return true;
}

User &User::operator>>(string &message)
{
  char tmp_buf[210] = { 0 };
  fd_set rfd;
  FD_ZERO(&rfd);
  FD_SET(this->socket_id(), &rfd);
  timeval tv = this->timeout();

  if(!select(this->socket_id() + 1, &rfd, NULL, NULL, &tv))
  {
    *this << "421 Connection timed out("
      << util.itos(config->get_user(this->user_name()).idle)
      << " seconds). Disconnected.";
    if(this->logged_in())
      logging->log_this(3, TYPE_AUTH, "User " + this->user_name() +
          ", disconnected due to timeout.");

    this->settings(FAILURE, B_SET);
  }
  else
  {
    if(this->receive(tmp_buf, 200) > 0)
      message = tmp_buf;
    else
      this->settings(FAILURE, B_SET);
  }
  return *this;
}

void kill_user(void *user)
{
  delete static_cast<User*>(user);
}
