/////////////////////////////////////////////////////////////////////
// $Id: Utilities.cpp,v 1.49 2005/04/06 18:06:28 klas Exp $
//
// Usage: Various functions that doesn't belong anywhere else.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define UTILITIES_CPP
#include "Utilities.h"

class Utilities util;

// function to remove line break at the end of a string
void Utilities::strip_line_breaks(string &victim)
{
  size_t last = victim.find_last_of('\n');
  if(last >= 0 && last < victim.size())
    victim = victim.erase(last);
}

// function to clean a string from:
//      leading spaces
//      trailing spaces
//      spaces between words
// returns number of spaces left in the string
int Utilities::clean_string(string &victim)
{
  int spaces = 0;
  
  // removing leading spaces...
  while(isspace(victim[0]))
    victim.erase(0, 1);

  // removing trailing spaces...
  while(isspace(victim[victim.size() - 1]))
    victim.erase(victim.size() - 1, 1);
  
  for(string::iterator victim_it = victim.begin(); victim_it != victim.end();
      ++victim_it)
  {
    if(isspace(*victim_it) || !isascii(*victim_it))
    {
      if(*victim_it == ' ')
        spaces++;
      if((isspace(*(victim_it - 1))) || !isascii(*victim_it))
      {
        if(*victim_it == ' ')
          spaces--;
        victim_it = victim.erase(victim_it);
        --victim_it;
      }
    }
  }
  return(spaces);
}

// function to convert an int to a string
string Utilities::itos(long long int num)
{
  string message;
  char buf[12] = {0};

  snprintf(buf, 12, "%lld", num);
  message = buf;

  return(message);
}

// function to parse a numeric permission (0755) to letters (-rwxr-xr-x)
string Utilities::parse_perms_ls(mode_t perm)
{
  string attr;

  attr = "----------";

  if(S_ISDIR(perm))
    attr[0] = 'd';
  else if(S_ISLNK(perm))
    attr[0] = 'l';
  else if(S_ISCHR(perm))
    attr[0] = 'c';
  else if(S_ISBLK(perm))
    attr[0] = 'b';
  else if(S_ISFIFO(perm))
    attr[0] = 'p';
  else if(S_ISSOCK(perm))
    attr[0] = 's';

  if(is_set(perm, S_IRUSR)) attr[1] = 'r';
  if(is_set(perm, S_IWUSR)) attr[2] = 'w';
  if(is_set(perm, S_IXUSR)) {
    if(is_set(perm, S_ISUID)) attr[3] = 's';
    else attr[3] = 'x';
  } else if(is_set(perm, S_ISUID)) attr[3] = 'S';

  if(is_set(perm, S_IRGRP)) attr[4] = 'r';
  if(is_set(perm, S_IWGRP)) attr[5] = 'w';
  if(is_set(perm, S_IXGRP)) {
    if(is_set(perm, S_ISGID)) attr[6] = 's';
    else attr[6] = 'x';
  } else if(is_set(perm, S_ISGID)) attr[6] = 'S';

  if(is_set(perm, S_IROTH)) attr[7] = 'r';
  if(is_set(perm, S_IWOTH)) attr[8] = 'w';
  if(is_set(perm, S_IXOTH)) {
    if(is_set(perm, S_ISVTX)) attr[9] = 't';
    else attr[9] = 'x';
  } else if(is_set(perm, S_ISVTX)) attr[9] = 'T';

  return(attr);
}

string Utilities::parse_perms_num(mode_t perm)
{
  string attr;

  attr = "0000";

  if(is_set(perm, S_ISUID))
    attr[0] += 4;
  if(is_set(perm, S_ISGID))
    attr[0] += 2;
  if(is_set(perm, S_ISVTX))
    attr[0]++;
  if(is_set(perm, S_IRUSR))
    attr[1] += 4;
  if(is_set(perm, S_IWUSR))
    attr[1] += 2;
  if(is_set(perm, S_IXUSR))
    attr[1]++;

  if(is_set(perm, S_IRGRP))
    attr[2] += 4;
  if(is_set(perm, S_IWGRP))
    attr[2] += 2;
  if(is_set(perm, S_IXGRP))
    attr[2]++;

  if(is_set(perm, S_IROTH))
    attr[3] += 4;
  if(is_set(perm, S_IWOTH))
    attr[3] += 2;
  if(is_set(perm, S_IXOTH))
    attr[3]++;

  return(attr);
}


int Utilities::time_string(string &victim, string format,
    size_t len, time_t timep)
{
  char *timestr = new char[len + 1];
  struct tm now;
  memset(timestr, 0, len + 1);

  if(timep == 0)
    timep = time(NULL);
  localtime_r(&timep, &now);
  int ret = strftime(timestr, len, format.c_str(), &now);
  victim = timestr;

  delete[] timestr;
  return(ret);
}
