/////////////////////////////////////////////////////////////////////
//  $Id: babyftpd.cpp,v 1.138 2005/08/23 19:34:02 klas Exp $
//
// Usage: main function, commandline parsing and function that
//        on the main socket.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#define BABYFTPD_CPP
#define NETINCLS
#define BABY_IO
#define BABY_CORE
#include "babyftpd.h"

int main(int argc, char **argv)
{
  pid_t pid;

  logging = new class Log;
  config = new class Configuration;
  config->parse_command_line(argc, argv);
  // we send the start message for logging here so that starts our log.
  logging->log_this(3, TYPE_INFO, config->general.welcome_msg + " started.");
  // the config file variable exists in the Configuration class :p
  if(config->parse_config_file() == false)
    cout << "Failed to open the config file (" << config->conf_file <<
      "), check if it exists + permissions." << endl;
  // if the conffile changes variables from commandline we reset them.
  config->parse_command_line(argc, argv);

  if(config->general.no_daemon == false)
    pid = fork();
  else
    pid = 0;
  
  switch(pid)
  {
    case -1:
      // error
      exit(1);
      break;
      
    case 0:
      // child
      int new_pid;
      setsid();
      if(config->general.no_daemon == false)
        new_pid = fork();
      else
        new_pid = 0;

      switch(new_pid)
      {
        case -1:
          exit(0);
          break;

        case 0:
          break;

        default:
          _exit(0);
      }

      chdir("/");
      umask(0);

      close(STDIN_FILENO);
      close(STDOUT_FILENO);
      close(STDERR_FILENO);

      if(!config->general.pid_file.empty())
      {
        ofstream pidstream;
        pidstream.open(config->general.pid_file.c_str(), ofstream::out);
        if(!pidstream)
          logging->log_this(2, TYPE_INFO, "Failed to open pid_file " +
              config->general.pid_file + " for writing. continuing.");
        else
        {
          pidstream << getpid();
          pidstream.close();
        }
      }

      // we can't spawn the logging thread until after we are daemonized
      pthread_create(&logging_thread, NULL, log_wrapper, NULL);

      pthread_key_create(&my_user, kill_user);
      pthread_key_create(&my_data, kill_data);

      pthread_rwlock_init(&user_list_lock, NULL);
      sem_init(&cleanup_list_lock, 0, 1);
      sem_init(&comm.pam_lock, 0, 1);

      signal(SIGTERM, sig_handler);
      signal(SIGINT, sig_handler);
      signal(SIGHUP, sig_handler);
      signal(SIGPIPE, sig_handler);

      main_thread = pthread_self();

#ifdef USE_TLS
#ifndef OLD_GCRYPT
      gcry_control(GCRYCTL_SET_THREAD_CBS, &gcry_threads_pthread);
#endif // OLD_GCRYPT
      int ret;
      pthread_rwlock_init(&cache_lock, NULL);
      gnutls_global_init();

      gnutls_certificate_allocate_credentials(&x509_cred);
      ret = gnutls_certificate_set_x509_trust_file(x509_cred,
          config->general.tls_ca.c_str(), GNUTLS_X509_FMT_PEM);
      if(ret < 0)
        logging->log_this(2, TYPE_INFO, gnutls_strerror(ret));

      ret = gnutls_certificate_set_x509_crl_file(x509_cred,
          config->general.tls_crl.c_str(), GNUTLS_X509_FMT_PEM);
      if(ret < 0)
        logging->log_this(2, TYPE_INFO, gnutls_strerror(ret));

      ret = gnutls_certificate_set_x509_key_file(x509_cred,
          config->general.tls_cert.c_str(), config->general.tls_key.c_str(),
          GNUTLS_X509_FMT_PEM);
      if(ret < 0)
        logging->log_this(2, TYPE_INFO, gnutls_strerror(ret));

      gnutls_dh_params_init(&dh_params);

      gnutls_dh_params_generate2(dh_params, 1024);

      gnutls_certificate_set_dh_params(x509_cred, dh_params);
      dh_time = time(NULL);

#endif // USE_TLS
      handler = new Handler;
      listener();
      break;

    default:
      // parent
      _exit(0);
      break;
  }
}

bool listener(void)
{
  struct sockaddr_in address;
  struct timeval tv = {config->general.lifebeat_tick,0};

  int first_socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  address.sin_family = AF_INET;
  address.sin_addr.s_addr = INADDR_ANY;
  address.sin_port = htons(config->general.port);
  if(bind(first_socket, reinterpret_cast<struct sockaddr*>(&address),
        sizeof(address)) < 0)
  {
    logging->log_this(2, TYPE_INFO, "Port " +
        util.itos(config->general.port) + " already in use.");
    exit_baby(1, EXIT_FULL);
  }
  seteuid(config->general.user);
  setgid(config->general.group);
  listen(first_socket, LISTEN_LIMIT);
  while(true)
  {
    pthread_t thread_tmp;
    socklen_t socketlen = sizeof(struct sockaddr_in);
    fd_set sock_set;
    time_t before = time(NULL) - (config->general.lifebeat_tick - tv.tv_sec);

    FD_ZERO(&sock_set);
    FD_SET(first_socket, &sock_set);
    if(!select(first_socket + 1, &sock_set, NULL, NULL, &tv))
    {
      lifebeat();
      tv.tv_sec = config->general.lifebeat_tick; // important to reset lifebeat
      continue;
    }

    int *final_socket = new int;
    *final_socket = accept(first_socket,
        reinterpret_cast<struct sockaddr*>(&address), &socketlen);
    if(*final_socket != -1)
      pthread_create(&thread_tmp, NULL, user_wrapper, (final_socket));
    else
    {
      logging->log_this(3, TYPE_INFO, "Failed to accept new caller.");
      // if the socket is accepted and the thread created the pointer will
      // get deleted there after it is stored.
      delete final_socket;
    }

    tv.tv_sec = config->general.lifebeat_tick - (time(NULL) - before);
    if(tv.tv_sec < 0 || tv.tv_sec > config->general.lifebeat_tick)
      tv.tv_sec = 0;
  }
  close(first_socket);

  return(true);
}

void lifebeat(void)
{
  logging->log_this(4, TYPE_DEBUG, "lifebeat reached.");

  sem_wait(&cleanup_list_lock);
  while(!cleanup_list.empty())
  {
    logging->log_this(4, TYPE_DEBUG, "joining thread #" +
        util.itos(cleanup_list.back()));
    pthread_join(cleanup_list.back(), NULL);
    cleanup_list.pop_back();
  }
  sem_post(&cleanup_list_lock);

#ifdef USE_TLS
  if((time(NULL) - dh_time) > config->general.dh_regenerate)
  {
    gnutls_dh_params temp_dh_params;
    gnutls_dh_params_init(&temp_dh_params);
    gnutls_dh_params_generate2(temp_dh_params, 1024);
    gnutls_dh_params_cpy(dh_params, temp_dh_params);
    gnutls_dh_params_deinit(temp_dh_params);
    gnutls_certificate_set_dh_params(x509_cred, dh_params);
    dh_time = time(NULL);
  }
#endif // USE_TLS
}

static void *user_wrapper(void *socket_id)
{
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);

  class User *user = handler->initializer((*(static_cast<int*>(socket_id))));
  delete (static_cast<int*>(socket_id));
  pthread_setspecific(my_user, static_cast<void*>(user));

  handler->init(user);

  logging->log_this(3, TYPE_DEBUG, "closed connection from: " +
      user->host_name());
  pthread_exit(NULL); // the thread exits..., and takes the user with it..
}

static void *log_wrapper(void *not_used)
{
  pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
  logging->log_main();

  pthread_exit(NULL);
}


void sig_handler(int signal)
{
  switch(signal)
  {
    case SIGTERM:
    case SIGINT:
      logging->log_this(3, TYPE_INFO, "SIGTERM received.");

      logging->log_this(4, TYPE_DEBUG, "main thread. users: " +
          util.itos(user_list.size()));

      for(vector<User*>::iterator user_it = user_list.begin();
          user_it != user_list.end(); user_it++)
        pthread_cancel((*user_it)->thread_id());

      struct timeval times;
      times.tv_sec = 1;
      times.tv_usec = 0;
      select(0, NULL, NULL, NULL, &times);

      logging->log_this(3, TYPE_INFO, "main thread exiting.");
      exit_baby(1, EXIT_FULL);
      break;

    case SIGHUP:
      logging->log_this(3, TYPE_INFO,
          "SIGHUP received, reloading configuration");
      config->parse_config_file();
      break;

    case SIGPIPE:
      logging->log_this(3, TYPE_INFO,
          "SIGPIPE received in thread: " + util.itos(pthread_self()));
      if(!pthread_equal(pthread_self(), main_thread) ||
          pthread_equal(pthread_self(), logging_thread))
        pthread_exit(NULL); // if this is a user or data thread, bail.
      break;
  }
}

void exit_baby(int num, enum exits how)
{
  if(how == EXIT_FULL)
  {
    if(!config->general.pid_file.empty())
      unlink(config->general.pid_file.c_str());

    while(!cleanup_list.empty())
    {
      logging->log_this(4, TYPE_DEBUG, "joining thread #" +
          util.itos(cleanup_list.back()));
      pthread_join(cleanup_list.back(), NULL);
      cleanup_list.pop_back();
    }

    delete handler;
    pthread_key_delete(my_user);
    pthread_key_delete(my_data);

#ifdef USE_TLS
    // emptying and freeing session cache.
    for(vector<cache*>::iterator cache_it = cache_db.begin();
        cache_it != cache_db.end(); cache_it++)
    {
      gnutls_free((*cache_it)->key.data);
      gnutls_free((*cache_it)->data.data);
      delete(*cache_it);
      cache_db.erase(cache_it);
    }

    gnutls_dh_params_deinit(dh_params);
    gnutls_certificate_free_credentials(x509_cred);
    gnutls_global_deinit();
    pthread_rwlock_destroy(&cache_lock);
#endif // USE_TLS 
    sem_destroy(&comm.pam_lock);
    sem_destroy(&cleanup_list_lock);
    pthread_rwlock_destroy(&user_list_lock);
    pthread_cancel(logging_thread);
    pthread_join(logging_thread, NULL);
  }
  delete logging;

  delete config;
  exit(num);
}

#ifdef USE_TLS // code for session resuming

int tls_db_store(void *dbf, gnutls_datum key, gnutls_datum data)
{
  struct cache *temp_cache = new cache;

  temp_cache->key.data = static_cast<unsigned char*>(gnutls_malloc(key.size));
  temp_cache->key.size = key.size;
  memcpy(temp_cache->key.data, key.data, key.size);

  temp_cache->data.data =
    static_cast<unsigned char*>(gnutls_malloc(data.size));
  temp_cache->data.size = data.size;
  memcpy(temp_cache->data.data, data.data, data.size);

  pthread_rwlock_wrlock(&cache_lock);
  cache_db.push_back(temp_cache);
  pthread_rwlock_unlock(&cache_lock);
  return(0);
}

gnutls_datum tls_db_fetch(void *dbf, gnutls_datum key)
{
  gnutls_datum data = { NULL, 0 };

  pthread_rwlock_rdlock(&cache_lock);
  for(vector<cache*>::iterator cache_it = cache_db.begin();
      cache_it != cache_db.end(); cache_it++)
    if(memcmp(key.data, (*cache_it)->key.data, (*cache_it)->key.size) == 0)
    {
      data.size = (*cache_it)->data.size;
      data.data = static_cast<unsigned char*>(gnutls_malloc(data.size));
      memcpy(data.data, (*cache_it)->data.data, data.size);
      break;
    }
  pthread_rwlock_unlock(&cache_lock);

  return(data);
}

int tls_db_delete(void *dbf, gnutls_datum key)
{
  int ret = -1;
  pthread_rwlock_rdlock(&cache_lock);
  for(vector<cache*>::iterator cache_it = cache_db.begin();
      cache_it != cache_db.end(); cache_it++)
    if(memcmp(key.data, (*cache_it)->key.data, (*cache_it)->key.size) == 0)
    {
      pthread_rwlock_unlock(&cache_lock);
      pthread_rwlock_wrlock(&cache_lock);
      gnutls_free((*cache_it)->key.data);
      gnutls_free((*cache_it)->data.data);
      delete(*cache_it);
      cache_db.erase(cache_it);
      ret = 0;
      break;
    }
  pthread_rwlock_unlock(&cache_lock);

  return(ret);
}

#endif // USE_TLS
