/////////////////////////////////////////////////////////////////////
//  $Id: generic.h,v 1.98 2005/08/23 19:34:02 klas Exp $
//
// Usage: some defines that is important for correct inclusions of
//        system headers, and all system header includes.
//
//  A part of babyftpd, licensed under the GNU GPL, see the file
//    LICENSE for complete information.
/////////////////////////////////////////////////////////////////////

#ifndef GENERIC_H
#define GENERIC_H

#define __USE_GNU 1

#include "config.h"
#include <string>
#include <vector>
#include <semaphore.h>
#include <assert.h>
#include <errno.h>

#ifdef NETINCLS
#include <sys/socket.h>
#include <arpa/inet.h>
#endif // NETINCLS

#ifdef BABY_IO
#include <iostream>
#include <fstream>
using std::cout;
using std::endl;
using std::ifstream;
using std::ofstream;
#endif // BABY_IO

#ifdef BABY_CORE
#include <pthread.h>
#include <bitset>
#include <security/pam_appl.h>
#include <ext/hash_map>
#ifdef USE_TLS
#include <gnutls/gnutls.h>
#endif // USE_TLS
using std::bitset;
using __gnu_cxx::hash_map;
using __gnu_cxx::hash;
#endif // BABY_CORE

#ifdef UTILITIES_H
#include <sys/stat.h>
#endif // UTILITIES_H

#ifdef BABYFTPD_H
#include <csignal>
#include <sys/stat.h>
#ifdef USE_TLS
#ifndef OLD_GCRYPT
#include <gcrypt.h>
#endif // OLD_GCRYPT
#endif // USE_TLS
#endif // BABYFTPD_H

#ifdef HANDLER_H
#include <algorithm>
#endif // HANDLER_H

#ifdef CONFIGURATION_H
#include <pwd.h>
#include <grp.h>
#include <getopt.h>
#endif // CONFIGURATION_H

#ifdef COMMANDS_H
#include <pwd.h>
#include <sys/stat.h>
#endif // COMMANDS_H

#ifdef COMMANDS2_CPP
#include <grp.h>
#include <dirent.h>
#include <fcntl.h>
#include <algorithm>
#endif // COMMANDS2_CPP

#ifdef LOG_H
#include <syslog.h>
#endif // LOG_H

#ifdef CONNECTION_H
#ifdef USE_TLS
#include <gnutls/x509.h>
#endif // USE_TLS
#endif // CONNECTION_H

#ifdef USER_H
#include <pwd.h>
#endif // USER_H

#ifdef DATA_H
#include <fcntl.h>
#include <unistd.h>
#endif // DATA_H

using std::string;
using std::vector;

class Configuration;
class User;
class Data;
class Handler;
class Commands;
class Utilities;
class StringTokenizer;
class Log;
class Connection;

const int PWBUF_SIZE = 100;

// for exit_baby()
enum exits { EXIT_EARLY, EXIT_FULL };

#ifndef FTP_LOG
#define FTP_LOG DAEMON_LOG
#endif // FTP_LOG

// various macros follow here

// these three wrapper macros around pthread_rwlock_* should be used
//  in all places where there is any risk one thread locks a lock twice without
//  unlocking. note that the unlock must occur before the int declared in the
//  lock has gone out of scope.
#ifdef BABY_CORE
#define BABY_WRLOCK(lock_var)   \
  int baby_lock_err = pthread_rwlock_wrlock(&lock_var); \
  assert(baby_lock_err == 0 || baby_lock_err == EDEADLK);
#define BABY_RDLOCK(lock_var)   \
  int baby_lock_err = pthread_rwlock_rdlock(&lock_var); \
  assert(baby_lock_err == 0 || baby_lock_err == EDEADLK);
#define BABY_UNLOCK(lock_var)   \
  if(baby_lock_err == 0)  \
    pthread_rwlock_unlock(&lock_var);
#endif // BABY_CORE
#endif // GENERIC_H
